/*
Copyright (C) 2001 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/


// Stuff to handle Processor
#include "stdafx.h"

#include "Registers.h"					// For REG_?? defines
#include "RSP_HLE.h"
#include "CPU.h"
#include "RSP.h"
#include "Memory.h"
#include "Interrupt.h"
#include "ROMBuffer.h"

#include "Resources\resource.h"

#include "Utility\ResourceString.h"
#include "Utility\Synchroniser.h"

#include "daedalus_crc.h"
#include "DaedThread.h"
#include "DaedTiming.h"

#include "Debug\Debug.h"
#include "Debug\DBGConsole.h"

#include "DynaRec\DynaRec.h"

#include "OSHLE\ultra_R4300.h"
#include "OSHLE\OS.h"
#include "OSHLE\Patch.h"				// GetCorrectOp

#include "Interface\MainWindow.h"		// For MWM_STARTEMU etc

#include "R4300.h"
#include "R4300_Regs.h"

#include "Plugins\GraphicsPlugin.h"
#include "Plugins\AudioPlugin.h"
#include "Plugins\DaedalusGraphics.h"

#ifdef DAEDALUS_XB
#include "GraphicsXb\GraphicsContext.h"
#endif


#ifdef DAEDALUS_W32
#define DAED_AUDIO_THREAD_ENABLED
#endif

//#define DAED_PC_COUNT

static const bool REALLY_ENABLE_IT = true;

//*****************************************************************************
//
//*****************************************************************************

#ifdef DAEDALUS_BREAKPOINTS_ENABLED
std::vector< DBG_BreakPoint > g_BreakPoints;
#endif

u32		gCPUStuffToDo			= 0;
static bool	gCPURunning			= false;
CAudioPlugin * g_pAiPlugin		= NULL;

CPUEvent *			gpFirstCPUEvent( NULL );
static CPUEvent *	gpCPUEventPool( NULL );

static u64			gLastVITime( 0 );
static float		gFrequencyInverse( 0.0f );

// XXXX TODO: This value can have a very large impact on the framerate.
static const u32	MAX_VI_TIME_SAMPLES( 2 );

static float		g_fVITimeSamples[ MAX_VI_TIME_SAMPLES ];
static DWORD		g_dwCurrentVITimeSample( 0 );
static f32			gCurrentRate( 0.0f );
static u32			gVerticalInterrupts( 0 );
static f32			gTvRefreshFrequency( 1.0f / 60.0f );
static u32			VI_INTR_CYCLES( 625000 );

static s32			gCPUThreadHandle( daedalus::INVALID_THREAD_HANDLE );
static volatile bool gCPUThreadActive( false );
#ifdef DAED_AUDIO_THREAD_ENABLED
static s32			gAudioThreadHandle( daedalus::INVALID_THREAD_HANDLE );
#endif

//*****************************************************************************
//
//*****************************************************************************
static u32 __stdcall CPUThreadFunc();
#ifdef DAED_AUDIO_THREAD_ENABLED
static u32 __stdcall AudioThreadFunc();
#endif

template< bool DynaRec, bool RunRSP > static void CPU_Go();

static void	CPU_RemoveHeadEvent();

#ifdef DAEDALUS_ENABLE_DYNAREC
static void (* g_pCPUCore)() = CPU_Go< true, false >;
#else
static void (* g_pCPUCore)() = CPU_Go< false, false >;
#endif

// TODO - On CPU_Fini, delete linked lists of events!

#ifdef DAED_PC_COUNT
//*****************************************************************************
//
//*****************************************************************************
struct PCCount
{
	u32 Address;
	u32 Count;

	bool operator<(const PCCount & rhs) const
	{
		return Count < rhs.Count;
	}
};
//
//	Map of hits per address
//
std::map<u32, u32> g_PCCount;

//*****************************************************************************
//
//*****************************************************************************
static void ResetPCCount()
{
	g_PCCount.clear();
}

//*****************************************************************************
//
//*****************************************************************************
static void AddPCCount(u32 program_counter)
{
	g_PCCount[ program_counter ]++;
}

//*****************************************************************************
//
//*****************************************************************************
static void DisplayPCCount()
{
	//
	//	Create a vector of all execution addresses and order by count
	//
	std::vector<PCCount>		arr_counts;
	
	arr_counts.reserve( g_PCCount.size() );

	for(std::map<u32,u32>::iterator it = g_PCCount.begin(); it != g_PCCount.end(); ++it)
	{
		PCCount	count;
		count.Address = it->first;
		count.Count = it->second;
		arr_counts.push_back( count );
	}

	std::sort( arr_counts.begin(), arr_counts.end() );
	
	for(u32 i = 0; i < arr_counts.size(); ++i)
	{
		DBGConsole_Msg(0, "hits at 0x%08x were %d", arr_counts[i].Address, arr_counts[i].Count);
	}
}
#endif

//*****************************************************************************
//
//*****************************************************************************
void CPU_Reset( u32 new_pc )
{
	u32 i;
	u32 dwReg;
	u32 dwCPU;
	DBGConsole_Msg(0, "Resetting with PC of 0x%08x", new_pc);

	u64 frequency;

	if(NTiming::GetPreciseFrequency(&frequency))
	{
		gFrequencyInverse = 1.0f / f32( frequency );
	}
	else
	{
		gFrequencyInverse = 1.0f;
	}

	// Initialise the array - assume perfect timekeeping
	for (i = 0; i < MAX_VI_TIME_SAMPLES; i++)
	{
		g_fVITimeSamples[i] = gTvRefreshFrequency;
	}

	// Point to the first sample
	gCurrentRate = 0.0f;
	g_dwCurrentVITimeSample = 0;

	CPU_SetPC( new_pc );
	g_qwMultHI = 0;
	g_qwMultLO = 0;

	for (dwReg = 0; dwReg < 32; dwReg++)
	{
		gGPR[dwReg]._u64 = 0;
		for (dwCPU = 0; dwCPU < 3; dwCPU++)
			g_qwCPR[dwCPU][dwReg]._u64 = 0;
	}

	// Init TLBs:
	for (i = 0; i < 32; i++)
	{	
		g_TLBs[i].pfno = 0;	// Clear Valid bit;
		g_TLBs[i].pfne = 0;	// Clear Valid bit;

		g_TLBs[i].pagemask = 0;

		g_TLBs[i].g		   = 1;

		g_TLBs[i].mask     =  g_TLBs[i].pagemask | 0x00001FFF;
		g_TLBs[i].mask2    =  g_TLBs[i].mask>>1;
		g_TLBs[i].vpnmask  = ~g_TLBs[i].mask;
		g_TLBs[i].vpn2mask =  g_TLBs[i].vpnmask>>1;

		g_TLBs[i].pfnehi = ((g_TLBs[i].pfne<<TLBLO_PFNSHIFT) & g_TLBs[i].vpn2mask);
		g_TLBs[i].pfnohi = ((g_TLBs[i].pfno<<TLBLO_PFNSHIFT) & g_TLBs[i].vpn2mask);
		
		g_TLBs[i].addrcheck = 0;
		g_TLBs[i].lastaccessed = 0;
	}

	// From R4300 manual			
    g_qwCPR[0][C0_RAND]._u64		= 32-1;			// TLBENTRIES-1
    //g_qwCPR[0][C0_SR]._u64		= 0x70400004;	//*SR_FR |*/ SR_ERL | SR_CU2|SR_CU1|SR_CU0;
	R4300_SetSR(0x70400004);
    g_qwCPR[0][C0_PRID]._u64	= 0x00000b10;	// Was 0xb00 - test rom reports 0xb10!!
	g_qwCPR[0][C0_CONFIG]._u64	= 0x0006E463;	// 0x00066463;	
	g_qwCPR[0][C0_WIRED]._u64   = 0x0;	

	g_qwCCR[1][0]._u64 = 0x00000511;
	
	Memory_MI_SetRegister(MI_VERSION_REG, 0x02020102);
	Memory_SP_SetRegister(SP_STATUS_REG, SP_STATUS_HALT);			// SP is halted

	((u32 *)g_pMemoryBuffers[MEM_RI_REG])[3] = 1;					// RI_CONFIG_REG Skips most of init


	// Look for Game boot address jump (so we know when to patch)
	// This op re-patches itself to R4300_Special_JR when 
	// the jump is detected
	R4300SpecialInstruction[OP_JR] = R4300_Special_JR_CheckBootAddress;

	g_nDelay = NO_DELAY;
	gCPUStuffToDo = 0;
	gVerticalInterrupts = 0;

	// Clear event list:
	while (gpFirstCPUEvent != NULL)
	{
		CPUEvent * pVictim = gpFirstCPUEvent;

		// Point to next event
		gpFirstCPUEvent = pVictim->pNext;

		// Add to pool:
		pVictim->pNext = gpCPUEventPool;
		gpCPUEventPool = pVictim;
	}

	CPU_AddEvent(VI_INTR_CYCLES, CPU_EVENT_VBL);

#ifdef DAEDALUS_BREAKPOINTS_ENABLED
	g_BreakPoints.clear();
#endif

#ifdef DAED_PC_COUNT
	ResetPCCount();
#endif

	//
	// We could call CPU_SelectCore here, but we know the RSP is halted, and we don't want to set any flags
	//
#ifdef DAEDALUS_ENABLE_DYNAREC
	if (g_bUseDynarec && REALLY_ENABLE_IT)
	{
		g_pCPUCore = CPU_Go< true, false >;
	}
	else
#endif
	{
		g_pCPUCore = CPU_Go< false, false >;
	}

}


//*****************************************************************************
// Thread stuff
//*****************************************************************************
bool CPU_StartThread( char * p_failure_reason, u32 length )
{
	if (!RomBuffer::IsRomLoaded())
	{
		strncpy( p_failure_reason, CResourceString(IDS_NOROMLOADED), length );
		p_failure_reason[ length - 1 ] = '\0';
		return false;
	}

	// If the thread is already running, just return
	if (gCPUThreadHandle != daedalus::INVALID_THREAD_HANDLE)
		return true;

#ifdef DAED_PC_COUNT
	ResetPCCount();
#endif
	// Attempt to create the thread
	gCPURunning = true;

#ifdef DAED_AUDIO_THREAD_ENABLED
	gAudioThreadHandle = daedalus::CreateThread( AudioThreadFunc, true );
#endif


#ifdef DAEDALUS_PSP
	CPUThreadFunc();
	return true;
#else
	gCPUThreadHandle = daedalus::CreateThread( CPUThreadFunc, true );
	if (gCPUThreadHandle != daedalus::INVALID_THREAD_HANDLE)
	{
		return true;
	}
	else
	{
		strncpy( p_failure_reason, CResourceString(IDS_UNABLETOSTARTCPUTHREAD), length );
		p_failure_reason[ length - 1 ] = '\0';
		return false;
	}
#endif
}

//*****************************************************************************
//
//*****************************************************************************
static void WaitForCPUThreadTermination()
{
	if(gCPUThreadHandle != daedalus::INVALID_THREAD_HANDLE)
	{
		// Wait forever for it to finish. It will clear/close gCPUThreadHandle when it exits
		while(gCPUThreadActive && daedalus::WaitForThreadTermination(gCPUThreadHandle, 1000))
		{
			DBGConsole_Msg(0, "Waiting for CPU thread (0x%08x) to finish", gCPUThreadHandle);
		}

		DAEDALUS_ASSERT( !gCPUThreadActive, "How come the thread is still marked as active?" );

		daedalus::ReleaseThreadHandle( gCPUThreadHandle );
		gCPUThreadHandle = daedalus::INVALID_THREAD_HANDLE;

		DBGConsole_Msg(0, "CPU Thread finished");
	}
}

//*****************************************************************************
//
//*****************************************************************************
void CPU_StopThread()
{
	// If it's no running, just return silently
	if (gCPUThreadHandle == daedalus::INVALID_THREAD_HANDLE)
		return;

	// If it is running, we need to signal for it to stop
	CPU_AddJob(CPU_STOP_RUNNING);

	WaitForCPUThreadTermination();

#ifdef DAED_PC_COUNT
	DisplayPCCount();
#endif
}

// BEGIN ADDED BY Lkb - 20/jul/2001 - Savestate support
//*****************************************************************************
//
//*****************************************************************************
extern bool g_bStopOnSimpleState;

void CPU_RunUntilStateIsSimple()
{
	if(gCPUThreadHandle != daedalus::INVALID_THREAD_HANDLE)
	{
		CPU_StopThread();
	}

	BOOL rsp_halted = Memory_SP_GetRegister( SP_STATUS_REG ) & SP_STATUS_HALT;

	bool simple_state = rsp_halted && !RSP_HLE_IsTaskLoaded() && (g_nDelay == NO_DELAY);	// Not sure if we need to check RSP_HLE_IsTaskLoaded??
	
	if( !simple_state )
	{
		g_bStopOnSimpleState = true;
		TCHAR dummyReason[MAX_PATH];
		CPU_StartThread(dummyReason, MAX_PATH);
		WaitForCPUThreadTermination();
		g_bStopOnSimpleState = false;
	}
}
// END ADDED BY Lkb - 20/jul/2001 - Savestate support

#ifdef DAED_AUDIO_THREAD_ENABLED
//*****************************************************************************
//
//*****************************************************************************
u32 __stdcall AudioThreadFunc()
{
	BOOL bResult; 
	CAudioPlugin * pAiPlugin;

	DBGConsole_Msg(0, "Audio thread started");

	//
	// Close the plugin if it's currently running (shouldn't be?)
	//
	pAiPlugin = g_pAiPlugin;

	// Close the existing plugin if it's already opened
	if ( pAiPlugin )
	{
		g_pAiPlugin = NULL;
		pAiPlugin->CloseDll();
		pAiPlugin->Close();
		delete pAiPlugin;
		pAiPlugin = NULL;
	}

	//
	// Only try to init if an audio plugin has been specified
	//
	if (strlen(g_DaedalusConfig.szAiPluginFileName) > 0)
	{
		DBGConsole_Msg(0, "Initialising Audio Plugin [C%s]", g_DaedalusConfig.szAiPluginFileName);
		pAiPlugin = new CAudioPlugin(g_DaedalusConfig.szAiPluginFileName);

		bResult = FALSE;
		if ( pAiPlugin )
		{
			bResult = pAiPlugin->Open();
			if (bResult)
				bResult = pAiPlugin->Init();

			// Stupid hack to force plugins to create the sound buffer
			pAiPlugin->DacrateChanged(SYSTEM_NTSC);
		}

		if (!bResult)
		{
			DBGConsole_Msg(0, "Error loading audio plugin");
		}
	}

	g_pAiPlugin = pAiPlugin;


	daedalus::SetThreadPriority( gAudioThreadHandle, TP_TIME_CRITICAL );

#ifdef DAEDALUS_USE_EXCEPTIONS
	try 
#endif
	{
		// Audio plugin may die on us - in which case the wrapper will unload the plugin
		while ( g_pAiPlugin && gCPURunning )
		{
			// There's a pretty huge overhead calling the normal Update() function 
			// (i.e. going through the critsect and the try/except block.)
			//g_pAiPlugin->Update(TRUE);
			g_pAiPlugin->UpdateFast( TRUE );
		}
	}
#ifdef DAEDALUS_USE_EXCEPTIONS
	catch (...)
	{
		if ( g_DaedalusConfig.TrapExceptions )
		{
			CMainWindow::Get()->MessageBox(CResourceString(IDS_AUDIOTHREAD_EXCEPTION),
										   g_szDaedalusName,
										   MB_ICONEXCLAMATION|MB_OK);
		}
		else
			throw;
	}
#endif

	DBGConsole_Msg(0, "Audio thread stopping");
	{
		// Make a copy of the plugin, and set the global pointer to NULL; 
		// This stops other threads from trying to access the plugin
		// while we're in the process of shutting it down.
		pAiPlugin = g_pAiPlugin;

		if (pAiPlugin != NULL)
		{
			g_pAiPlugin = NULL;
			pAiPlugin->CloseDll();
			pAiPlugin->Close();
			delete pAiPlugin;
			pAiPlugin = NULL;
		}
	}

	//
	//	Ick - I think this should really be done from the main thread
	//
	daedalus::ReleaseThreadHandle( gAudioThreadHandle );
	gAudioThreadHandle = daedalus::INVALID_THREAD_HANDLE;

	return 0;
}
#endif // DAED_AUDIO_THREAD_ENABLED

//*****************************************************************************
//
//*****************************************************************************
u32 __stdcall CPUThreadFunc()
{
	gCPUThreadActive = true;

#if defined(DAEDALUS_W32)
	// From Lkb- avoid GFX exceptions when the RDP is configured for the ROM ucode 
	//CMainWindow::Get()->PostMessage( CMainWindow::MWM_STARTEMU, 0,0);
	CMainWindow::Get()->SendMessage( CMainWindow::MWM_STARTEMU, 0,0 );

	{
		DAEDALUS_ASSERT( !CGraphicsPlugin::IsAvailable(), "The graphics plugin should not be initialised at this point" );
		
		if (strlen(g_DaedalusConfig.szGfxPluginFileName) > 0)
		{
			BOOL bResult;
			DBGConsole_Msg(0, "Initialising Graphics Plugin [C%s]", g_DaedalusConfig.szGfxPluginFileName);

			CGraphicsPlugin::Create();

			bResult = CGraphicsPlugin::Get()->Open(g_DaedalusConfig.szGfxPluginFileName);
			if (bResult)
			{
				bResult = CGraphicsPlugin::Get()->Init();

				if (bResult)
				{
					CGraphicsPlugin::Get()->RomOpen();
					CGraphicsPlugin::Get()->ExecuteCommand( DAEDALUS_GFX_USENEWCOMBINER, (LPVOID)g_bDisplayNewCombiner );
					CGraphicsPlugin::Get()->ExecuteCommand( DAEDALUS_GFX_SET_DEBUG, CDebugConsole::Get() );
				}
			}

			if (!bResult)
			{
				DBGConsole_Msg(0, "Error loading graphics plugin");
				CGraphicsPlugin::Get()->Close();
				CGraphicsPlugin::Destroy();
			}
		}
	}
#elif defined(DAEDALUS_PSP)

	{

		BOOL bResult;
		DBGConsole_Msg(0, "Initialising Graphics Plugin [C%s]", g_DaedalusConfig.szGfxPluginFileName);

		bResult = Graphics_InitiateGFX();

		if (bResult)
		{
			Graphics_RomOpen();
			//Graphics_ExecuteCommand( "UseNewCombiner", (LPVOID*)&g_bDisplayNewCombiner );
		}
		else
		{
			DBGConsole_Msg(0, "Error loading graphics plugin");
		}

	}
#else

#endif

#ifdef DAEDALUS_USE_EXCEPTIONS
	try
#endif
	{
		while (gCPURunning)
		{
			// This function is either CPU_Go< x, true > or CPU_Go< x, false >
			// depending on whether the RSP is halted or not
			g_pCPUCore();
		}
	}
#ifdef DAEDALUS_USE_EXCEPTIONS
	catch (...)
	{
		if (g_DaedalusConfig.TrapExceptions)
		{
			CMainWindow::Get()->MessageBox(CResourceString(IDS_CPUTHREAD_EXCEPTION),
										   g_szDaedalusName,
										   MB_ICONEXCLAMATION|MB_OK);
			gCPURunning = false;
		}
		else
			throw;
	}
#endif

	// NB - this order is very critical. The main thread is polling on gCPUThreadActive after
	// a call to CPU_StopThread(), and so it can't process its message pump. This means that
	// any calls below that can potentially send a message to our main window will block indefinitely.
	gCPUThreadActive = false;

	// Update the screen. It's probably better handled elsewhere...
	DBGConsole_UpdateDisplay();

	if ( CGraphicsPlugin::IsAvailable() )
	{
		CGraphicsPlugin::Get()->RomClosed();
		CGraphicsPlugin::Get()->CloseDll();
#ifdef DAEDALUS_W32
		CGraphicsPlugin::Get()->Close();
		CGraphicsPlugin::Destroy();
#endif
	}

#ifdef DAEDALUS_W32
	// This causes a few problems with synchronisation if we call here
	// (it passes a message to the main window, but the main window's 
	// handling thread is busy waiting for this thread to exit). I've
	// put the call at the end of CPU_StopThread()
	//Main_ActivateList();
	CMainWindow::Get()->PostMessage(CMainWindow::MWM_ENDEMU, 0,0);
#endif

	return 0;
}

//*****************************************************************************
//
//*****************************************************************************
void CPU_DynarecEnable()
{
	g_bUseDynarec = TRUE;
	CPU_AddJob(CPU_CHANGE_CORE);
}

// ADDED BY Lkb - 20/jul/2001 - Savestate support
bool g_bStopOnSimpleState = false;

//*****************************************************************************
//
//*****************************************************************************
void CPU_SelectCore()
{
	bool run_rsp = (Memory_SP_GetRegister( SP_STATUS_REG ) & SP_STATUS_HALT) == 0;

#ifdef DAEDALUS_ENABLE_DYNAREC
	if (g_bUseDynarec && REALLY_ENABLE_IT)
	{
		if (run_rsp)
			g_pCPUCore = CPU_Go< true, true >;
		else
			g_pCPUCore = CPU_Go< true, false >;
	}
	else
#endif
	{
		if (run_rsp)
			g_pCPUCore = CPU_Go< false, true >;
		else
			g_pCPUCore = CPU_Go< false, false >;
	}

	bool simple_state = !run_rsp && !RSP_HLE_IsTaskLoaded() && (g_nDelay == NO_DELAY);

	if( g_bStopOnSimpleState && simple_state )
		CPU_AddJob( CPU_STOP_RUNNING );
	else
		CPU_AddJob( CPU_CHANGE_CORE );
}

//*****************************************************************************
//
//*****************************************************************************
void CPU_Halt( const char * reason )
{
	DBGConsole_Msg( 0, "CPU Halting: %s", reason );
	CPU_AddJob( CPU_STOP_RUNNING );
}

#ifdef DAEDALUS_BREAKPOINTS_ENABLED
//*****************************************************************************
//
//*****************************************************************************
void CPU_AddBreakPoint( u32 address )
{
	OpCode * pdwOp;

	// Force 4 byte alignment
	address &= 0xFFFFFFFC;

	if (!Memory_GetInternalReadAddress(address, (void**)&pdwOp))
	{
		DBGConsole_Msg(0, "Invalid Address for BreakPoint: 0x%08x", address);
	}
	else
	{
		DBG_BreakPoint bpt;
		DBGConsole_Msg(0, "[YInserting BreakPoint at 0x%08x]", address);
		
		bpt.mOriginalOp = *pdwOp;
		bpt.bEnabled = TRUE;
		bpt.bTemporaryDisable = FALSE;
		g_BreakPoints.push_back(bpt);

		pdwOp->op = OP_DBG_BKPT;
		pdwOp->bp_index = (g_BreakPoints.size() - 1);
	}
}
#endif

#ifdef DAEDALUS_BREAKPOINTS_ENABLED
//*****************************************************************************
//
//*****************************************************************************
void CPU_EnableBreakPoint( u32 address, bool enable )
{
	OpCode * pdwOp;

	// Force 4 byte alignment
	address &= 0xFFFFFFFC;

	if (!Memory_GetInternalReadAddress(address, (void**)&pdwOp))
	{
		DBGConsole_Msg(0, "Invalid Address for BreakPoint: 0x%08x", address);
	}
	else
	{
		OpCode op_code;
		DWORD dwBreakPoint;

		op_code = *pdwOp;

		if (op_code.op != OP_DBG_BKPT)
		{
			DBGConsole_Msg(0, "[YNo breakpoint is set at 0x%08x]", address);
			return;
		}

		// Entry is in lower 26 bits...
		dwBreakPoint = op_code.bp_index;

		if (dwBreakPoint < g_BreakPoints.size())
		{
			g_BreakPoints[dwBreakPoint].bEnabled = enable;
			// Alwyas disable
			g_BreakPoints[dwBreakPoint].bTemporaryDisable = FALSE;
		}
	}
}
#endif

//*****************************************************************************
//
//*****************************************************************************
static void CPU_HANDLE_COUNT_INTERRUPT()
{
	CPUEvent * pEvent = gpFirstCPUEvent;
	if (pEvent == NULL)
		return;

	// Remove this event from the list, and place in the pool. 
	// The pointer is still valid afterwards
	CPU_RemoveHeadEvent();
	

	switch (pEvent->mEventType)
	{
	case CPU_EVENT_VBL:
		{
			u64		now;
			u64		elapsed_ticks;
			float fElapsed; 
			//float fSleepTime;
			float fSumOfVITimeSamples;

			u32		vertical_sync_reg( *VI_REG_ADDRESS(VI_V_SYNC_REG) );
			if (vertical_sync_reg == 0)
			{
				VI_INTR_CYCLES = 625000;
			}
			else
			{
				VI_INTR_CYCLES = (vertical_sync_reg+1) * 1500;
			}

			// Add another Interrupt at the next time:
			CPU_AddEvent(VI_INTR_CYCLES, CPU_EVENT_VBL);

			gVerticalInterrupts++;

			{
				float fCurrentRate;
				// dwSleepTime;

				// We're aiming for 60 vis/sec
				// This is a bit of a hack - busy wait until we're exactly on schedule
				do {

					// Check how long since last vi
					NTiming::GetPreciseTime(&now);

					elapsed_ticks = now - gLastVITime;
					fElapsed = f32( elapsed_ticks ) * gFrequencyInverse;

					// We could do a "rolling" sum (subtract the item to be replaced, add the
					// new values) but this will slowy degrade through fp inaccuracy
					g_fVITimeSamples[g_dwCurrentVITimeSample] = fElapsed;

					// Update the running total - subtract the value we removed, and add the new value
					fSumOfVITimeSamples = 0.0f;
					for (u32 i = 0; i < MAX_VI_TIME_SAMPLES; i++)
						fSumOfVITimeSamples += g_fVITimeSamples[i];
					fCurrentRate = fSumOfVITimeSamples / (float)MAX_VI_TIME_SAMPLES;

					// If we need to wait over 1 millisec, use Sleep()?
					/*fSleepTime = gTvRefreshFrequency - fCurrentRate;
					if (fSleepTime > 0)
					{
						dwSleepTime = (DWORD)(fSleepTime * 1000.0f);

						//DBGConsole_Msg(0, "Sleeping for %f -> %d", fSleepTime, (DWORD)(fSleepTime * 1000));
						if (dwSleepTime > 0)
							Sleep(dwSleepTime);
					}*/


				}
				while (g_bSpeedSync && fCurrentRate < gTvRefreshFrequency);

				//DBGConsole_Msg( 0, DSPrintf( "Last frame took %.3fms\n", fCurrentRate * 1000.0f ) );
				gCurrentRate = fCurrentRate;

				gLastVITime = now;
				// Point to next sample, and loop round if necessary
				g_dwCurrentVITimeSample = (g_dwCurrentVITimeSample + 1) % MAX_VI_TIME_SAMPLES;
			}

			Memory_MI_SetRegisterBits(MI_INTR_REG, MI_INTR_VI);
			R4300_Interrupt_UpdateCause3();

			if (CGraphicsPlugin::IsAvailable())
			{
				CGraphicsPlugin::Get()->UpdateScreen();
			}
		}
		break;
	case CPU_EVENT_COMPARE:
		{
			//DBGConsole_Msg(0, "Compare Interrupt 0x%08x", (u32)g_qwCPR[0][C0_COMPARE]);
			//CPU_Halt("Compare");
			g_qwCPR[0][C0_CAUSE]._u64 |= CAUSE_IP8;
			CPU_AddJob( CPU_CHECK_INTERRUPTS );
		}
		break;
	default:
		NODEFAULT;
	}
}


//*****************************************************************************
//
//*****************************************************************************
void CPU_SetCompare(u64 qwNewValue)
{
	g_qwCPR[0][C0_CAUSE]._u64 &= ~CAUSE_IP8;

	DPF( DEBUG_REGS, DSPrintf( "COMPARE set to 0x%08x.", (u32)qwNewValue ) );
	//DBGConsole_Msg(0, "COMPARE set to 0x%08x Count is 0x%08x.", (u32)qwNewValue, (u32)g_qwCPR[0][C0_COUNT]);

	// Add an event for this compare:
	if (qwNewValue == g_qwCPR[0][C0_COMPARE]._u64)
	{
		//	DBGConsole_Msg(0, "Clear");
	}
	else
	{
		if ((u32)qwNewValue > g_qwCPR[0][C0_COUNT]._u32[0])
		{
			CPU_AddEvent((u32)qwNewValue - g_qwCPR[0][C0_COUNT]._u32[0], CPU_EVENT_COMPARE);									
		}
		else if (qwNewValue != 0)
		{
			//DBGConsole_Msg(0, "COMPARE set to 0x%08x%08x < Count is 0x%08x%08x.",
			//	(u32)(qwNewValue>>32), (u32)qwNewValue,
			//	(u32)(g_qwCPR[0][C0_COUNT]>>32), g_qwCPR[0][C0_COUNT]._u32[0]);

			CPU_AddEvent((u32)qwNewValue - g_qwCPR[0][C0_COUNT]._u32[0], CPU_EVENT_COMPARE);									
			//DBGConsole_Msg(0, "0x%08x", (u32)qwNewValue - g_qwCPR[0][C0_COUNT]._u32[0]);
		}
		g_qwCPR[0][C0_COMPARE]._u64 = qwNewValue;
	}


}

//*****************************************************************************
//
//*****************************************************************************
u32	CPU_ProduceRegisterHash()
{
	u32	hash( 0 );

	if ( DAED_SYNC_MASK & DAED_SYNC_REG_GPR )
	{
		hash ^= daedalus_crc32( 0, (u8*)&(gGPR[0]), sizeof( gGPR ) );
	}

	if ( DAED_SYNC_MASK & DAED_SYNC_REG_CPU0 )
	{
		hash ^= daedalus_crc32( 0, (u8*)&(g_qwCPR[0][0]), sizeof( g_qwCPR[0] ) );
	}
	if ( DAED_SYNC_MASK & DAED_SYNC_REG_CCR0 )
	{
		hash ^= daedalus_crc32( 0, (u8*)&(g_qwCCR[0][0]), sizeof( g_qwCCR[0] ) );
	}

	if ( DAED_SYNC_MASK & DAED_SYNC_REG_CPU1 )
	{
		hash ^= daedalus_crc32( 0, (u8*)&(g_qwCPR[1][0]), sizeof( g_qwCPR[1] ) );
	}
	if ( DAED_SYNC_MASK & DAED_SYNC_REG_CCR1 )
	{
		hash ^= daedalus_crc32( 0, (u8*)&(g_qwCCR[1][0]), sizeof( g_qwCCR[1] ) );
	}

	return hash;
}

//*****************************************************************************
//	Execute a single MIPS op. The conditionals for the templated arguments
//	are completely optimised away by the compiler.
//	
//	DynaRec:		Run this function with dynarec enabled
//	TranslateOp:	Use this to translate breakpoints/patches to original op
//					before execution.
//*****************************************************************************
template< bool DynaRec, bool TranslateOp > __forceinline void CPU_EXECUTE_OP()
{
	gCurrentOp = *(OpCode *)g_pPCMemBase;

	if ( TranslateOp )
	{
	#ifdef DAEDALUS_BREAKPOINTS_ENABLED
		// Handle breakpoints correctly
		if (gCurrentOp.op == OP_DBG_BKPT)
		{
			// Turn temporary disable on to allow instr to be processed
			// Entry is in lower 26 bits...
			u32	breakpoint( gCurrentOp.bp_index );

			if ( breakpoint < g_BreakPoints.size() )
			{
				if (g_BreakPoints[ breakpoint ].bEnabled)
				{
					g_BreakPoints[ breakpoint ].bTemporaryDisable = true;
				}
			}
		}
		else
	#endif
		{
			gCurrentOp = GetCorrectOp( gCurrentOp );
		}
	}

	R4300Instruction[ gCurrentOp.op ]( );

	SYNCH_POINT( DAED_SYNC_REGS, CPU_ProduceRegisterHash(), "Registers don't match" );

	// Increment count register
	*(u32 *)&g_qwCPR[0][C0_COUNT] = *(u32 *)&g_qwCPR[0][C0_COUNT]+COUNTER_INCREMENT_PER_OP;

	gpFirstCPUEvent->mCount -= COUNTER_INCREMENT_PER_OP;
	if (gpFirstCPUEvent->mCount <= 0)
	{
		CPU_HANDLE_COUNT_INTERRUPT();
	}

	switch (g_nDelay)
	{
		case DO_DELAY:
			// We've got a delayed instruction to execute. Increment
			// PC as normal, so that subsequent instruction is executed
			INCREMENT_PC();
			g_nDelay = EXEC_DELAY;

			break;
		case EXEC_DELAY:
			// We've just executed the delayed instr. Now carry out jump as stored in g_dwNewPC;
			CPU_SetPC(g_dwNewPC);
			g_nDelay = NO_DELAY;

			if ( DynaRec )
			{
				// Only attempt to recompile *after* a jump target!
				gCurrentOp = *(OpCode *)g_pPCMemBase;
				if (!OP_IS_A_HACK( gCurrentOp._u32 ))
					DynaRec_CompileCode(g_dwPC);
			}
			break;
		case NO_DELAY:
			// Normal operation - just increment the PC
			INCREMENT_PC();

			if ( DynaRec )
			{
				gCurrentOp = *(OpCode *)g_pPCMemBase;
				if (!OP_IS_A_HACK( gCurrentOp._u32 ))
					DynaRec_CompileCode(g_dwPC);
			}

			break;
		default:	// MSVC extension - the default will never be reached
			NODEFAULT;
	}	
}

//*****************************************************************************
// Hacky function to use when debugging
//*****************************************************************************
void CPU_Skip()
{
	if (gCPURunning || gCPUThreadHandle != daedalus::INVALID_THREAD_HANDLE)
	{
		DBGConsole_Msg(0, "Already Running");
		return;
	}

	INCREMENT_PC();
}

//*****************************************************************************
//
//*****************************************************************************
void CPU_Step()
{
	if (gCPURunning || gCPUThreadHandle != daedalus::INVALID_THREAD_HANDLE)
	{
		DBGConsole_Msg(0, "Already Running");
		return;
	}

	// We do this in a slightly different order to ensure that
	// any interrupts are taken care of before we execute an op
	if (gCPUStuffToDo)
	{	
		// Process Interrupts/Exceptions on a priority basis
		// Call most likely first!
		if (gCPUStuffToDo & CPU_CHECK_INTERRUPTS)
		{
			R4300_Handle_Interrupt();
			gCPUStuffToDo &= ~CPU_CHECK_INTERRUPTS;
		}
		else if (gCPUStuffToDo & CPU_CHECK_EXCEPTIONS)
		{
			R4300_Handle_Exception();
			gCPUStuffToDo &= ~CPU_CHECK_EXCEPTIONS;
		}
		else if (gCPUStuffToDo & CPU_STOP_RUNNING)
		{
			gCPUStuffToDo &= ~CPU_STOP_RUNNING;
			gCPURunning = false;
		}
		else if (gCPUStuffToDo & CPU_CHANGE_CORE)
		{
			gCPUStuffToDo &= ~CPU_CHANGE_CORE;
		}
		// Clear gCPUStuffToDo?
	}

	CPU_EXECUTE_OP< false, true >();

	bool rsp_halted( (Memory_SP_GetRegister( SP_STATUS_REG ) & SP_STATUS_HALT) != 0 );

	if ( !rsp_halted )
	{
		RSP_Step();
	}

	DBGConsole_UpdateDisplay();
}


//*****************************************************************************
// Keep executing instructions until there are other tasks to do (i.e. gCPUStuffToDo is set)
// Process these tasks and loop
//*****************************************************************************
template < bool DynaRec, bool RunRSP > void CPU_Go()
{
	while (gCPURunning)
	{
		//
		// Keep executing ops as long as there's nothing to do
		//
		while (gCPUStuffToDo == 0)
		{	
			if ( RunRSP )
			{
				RSP_Step();							// At this point we know that the RSP is running. RSPStep() asserts that it is not halted
			}
			CPU_EXECUTE_OP< DynaRec, false >();
		}
		
		//
		// Handle whatever it was that cause us to stop
		//
		if (gCPUStuffToDo & CPU_CHECK_INTERRUPTS)
		{
			R4300_Handle_Interrupt();
			gCPUStuffToDo &= ~CPU_CHECK_INTERRUPTS;
		}
		else if (gCPUStuffToDo & CPU_CHECK_EXCEPTIONS)
		{
			R4300_Handle_Exception();
			gCPUStuffToDo &= ~CPU_CHECK_EXCEPTIONS;
		}
		else if (gCPUStuffToDo & CPU_STOP_RUNNING)
		{
			gCPUStuffToDo &= ~CPU_STOP_RUNNING;
			gCPURunning = false;
		}
		else if (gCPUStuffToDo & CPU_CHANGE_CORE)
		{
			// Break out of this loop. The new core is selected automatically
			gCPUStuffToDo &= ~CPU_CHANGE_CORE;
			break;
		}
	}
}

//*****************************************************************************
//
//*****************************************************************************
void CPU_AddEvent( u32 count, ECPUEventType event )
{
	if (count == 0)
	{
//		DBGConsole_Msg(0, "CPU: Attempt to add event with 0 time");
		return;
	}

	// Insert event at Count cycles into list
	CPUEvent * pNewEvent = gpCPUEventPool;
	if (pNewEvent)
	{
		gpCPUEventPool = pNewEvent->pNext;
	}
	else
	{
		pNewEvent = new CPUEvent;
		// Handle failures?
		if (pNewEvent == NULL)
			return;
	}

	pNewEvent->mCount = count;
	pNewEvent->mEventType = event;
	

	CPUEvent * pCheckEvent = gpFirstCPUEvent;
	CPUEvent * pPrev = NULL;

	while (pCheckEvent != NULL &&
		   pNewEvent->mCount > pCheckEvent->mCount)
	{
		// Decrease time by time waiter for this event
		pNewEvent->mCount -= pCheckEvent->mCount;

		// Check the next event
		pPrev = pCheckEvent;
		pCheckEvent = pCheckEvent->pNext;
	}

	// pCheckEvent points to an event that occurs AFTER us, so we need to update it:
	if (pCheckEvent != NULL)
	{
		pCheckEvent->mCount -= pNewEvent->mCount;

		// Link into this list
		pNewEvent->pPrev = pCheckEvent->pPrev;
		pNewEvent->pNext = pCheckEvent;

		if (pNewEvent->pPrev != NULL)
			pNewEvent->pPrev->pNext = pNewEvent;
		else
			gpFirstCPUEvent = pNewEvent;
		
		pNewEvent->pNext->pPrev = pNewEvent;
	}
	else if (pPrev != NULL)
	{
		// We've been inserted at the end of the list: Insert us after pPrev
		pNewEvent->pPrev = pPrev;
		pNewEvent->pNext = NULL;

		pNewEvent->pPrev->pNext = pNewEvent;
	}
	else
	{
		// We must be the first event in the list
		pNewEvent->pPrev = NULL;
		pNewEvent->pNext = NULL;
		gpFirstCPUEvent = pNewEvent;
	}
}

//*****************************************************************************
// Remove the event from the head of the event list, and place in the pool queue
// Ignores count of item being removed!
//*****************************************************************************
void CPU_RemoveHeadEvent()
{
	if (gpFirstCPUEvent != NULL)
	{
		CPUEvent * pEvent = gpFirstCPUEvent;

		if (pEvent->pNext != NULL)
			pEvent->pNext->pPrev = NULL;

		gpFirstCPUEvent = pEvent->pNext;

		// Add to head of pool:
		pEvent->pNext = gpCPUEventPool;
		gpCPUEventPool = pEvent;
	}
}

//*****************************************************************************
//
//*****************************************************************************
CPUEvent* CPU_GetVideoInterruptEvent()
{
	CPUEvent* p;
	for(p = gpFirstCPUEvent; p != NULL; p = p->pNext)
	{
		if(p->mEventType == CPU_EVENT_VBL)
			return p;
	}

	return NULL;
}

//*****************************************************************************
//
//*****************************************************************************
void	CPU_SetTvFrequency( f32 frequency )
{
	gTvRefreshFrequency = 1.0f / frequency;
}

//*****************************************************************************
//	Return true if the CPU is running
//*****************************************************************************
bool	CPU_IsRunning()
{
	return gCPURunning;
}

//*****************************************************************************
//	Get the current sync
//*****************************************************************************
f32	CPU_GetSync()
{
	return gTvRefreshFrequency / gCurrentRate;
}

//*****************************************************************************
//	Return the total number of vi interrupts
//*****************************************************************************
u32	CPU_GetVerticalInterruptCount()
{
	return gVerticalInterrupts;
}
