/*

  Copyright (C) 2001 StrmnNrmn

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// TODO: Add a CriticalSection to stop multiple threads using the 
// Unique objects simultaneously

#include "stdafx.h"
#include "DaedalusGraphics.h"
#include "PSPRenderer.h"
#include "Interface\IniFile.h"

#include "GraphicsContext.h"
#include "VideoMemoryManager.h"
#include "Plugins\GraphicsPlugin_Spec.h"

#include "Utility\Profiler.h"
#include "TextureCache.h"

#include "Interface\IniFile.h"
#include "DLParser.h"

#include <pspctrl.h>
#include <pspdisplay.h>
#include <pspdebug.h>
#include <pspgu.h>

#define DAEDALUSGRAPHICS_API __declspec(dllexport)

#define DAEDALUS_GFX_PLUGIN_VERSION		0x0001

DWORD g_dwDaedalusVersion = 0;

DWORD g_dwViWidth = 320;
DWORD g_dwViHeight = 240;

#define OS_TV_PAL		0
#define OS_TV_NTSC		1
#define OS_TV_MPAL		2

static bool gDropAllTextures = false;
 bool gDebugDisplayList = false;

CCritSect g_CritialSection;


bool GraphicsInitialise()
{
	if(!PSPRenderer::Create())
		return false;

	if(!CTextureCache::Create())
		return false;

	return true;
}


/******************************************************************
  Function: InitiateGFX
  Purpose:  This function is called when the DLL is started to give
            information from the emulator that the n64 graphics
			uses. This is not called from the emulation thread.
  Input:    Gfx_Info is passed to this function which is defined
            above.
  Output:   TRUE on success
            FALSE on failure to initialise
             
  ** note on interrupts **:
  To generate an interrupt set the appropriate bit in MI_INTR_REG
  and then call the function CheckInterrupts to tell the emulator
  that there is a waiting interrupt.
*******************************************************************/ 
BOOL Graphics_InitiateGFX ()
{
	g_CritialSection.Lock();
	
	GraphicsInitialise();

	if (!DLParser_Init()) 
	{
		g_CritialSection.Unlock();
		return FALSE;
	}

	g_CritialSection.Unlock();
	return TRUE;
}

/******************************************************************
  Function: CloseDLL
  Purpose:  This function is called when the emulator is closing
            down allowing the dll to de-initialise.
  input:    none
  output:   none
*******************************************************************/ 
DAEDALUSGRAPHICS_API void __cdecl CloseDLL (void)
{
	// Do nothing - do processing in DLL_PROCESS_DETACH
	g_CritialSection.Lock();
	g_CritialSection.Unlock();
}

#ifdef DAEDALUS_DEBUG_DISPLAYLIST

extern float	GetElapsedTime();
extern void		ResetElapsedTime();


/******************************************************************

*******************************************************************/ 
static void Graphics_DebugDList()
{
	//
	//	Enter the debug menu as soon as select is newly pressed
	//
    SceCtrlData pad;

	static u32 oldButtons = 0;

	sceCtrlReadBufferPositive(&pad, 1);

	oldButtons = pad.Buttons;

	bool menu_button_pressed( false );

	u64		freq;
	NTiming::GetPreciseFrequency( &freq );
	float freq_inv = 1.0f / f32( freq );

	PSPRenderer::Get()->SetRecordCombinerStates( true );

	bool	show_combiner_states( false );

	ResetElapsedTime();

	// Remain paused until the Select button is pressed again
	while(!menu_button_pressed)
	{
		//guSwapBuffersBehaviour( PSP_DISPLAY_SETBUF_IMMEDIATE );

		u64			time_before;
		NTiming::GetPreciseTime( &time_before );

		//
		//	Re-render the current frame
		//
		DLParser_Process();

		u64			time_after;
		NTiming::GetPreciseTime( &time_after );

		//
		//	Figure out how long the last frame took
		//
		u64			elapsed_ticks( time_after - time_before );
		float		elapsed_ms( f32(elapsed_ticks) * 1000.0f * freq_inv );
		float		framerate( 0.0f );
		if(elapsed_ms > 0)
		{
			framerate = 1000.0f / elapsed_ms;
		}

		Daedalus_ForceDebugScreenToFront();
		pspDebugScreenSetXY(0, 0);
		pspDebugScreenPrintf( "Dlist took %dms (%fhz)\n", s32(elapsed_ms), framerate );
		//pspDebugScreenPrintf( "Vertices - In: %d, Out: %d\n", gTesselatorVerticesIn, gTesselatorVerticesOut );

		//CTextureCache::Get()->DisplayStats();
		//pspDebugScreenPrintf( "Software clipping %s\n", gSoftwareClipping ? "enabled" : "disabled" );

		sceCtrlReadBufferPositive(&pad, 1);

		float actual_elapsed_time( GetElapsedTime() );

		{

			extern float DECAL_Z_OFFSET;
			const float DECAL_Z_CHANGE_PER_SECOND = 1;
			static const s32	STICK_DEADZONE = 20;

			pspDebugScreenPrintf( "ZDecal offset is %f\n", DECAL_Z_OFFSET );


			s32		stick_y( pad.Ly - 128 );

			if(stick_y >= -STICK_DEADZONE && stick_y <= STICK_DEADZONE)
			{
				stick_y = 0;
			}

			float pad_y( float(stick_y) / 128.0f );

			DECAL_Z_OFFSET += pad_y * DECAL_Z_CHANGE_PER_SECOND * actual_elapsed_time; 


		}

		if(show_combiner_states)
		{
			static u32	selected_idx( 0 );

			const std::set< u64 > & 	combiner_states( PSPRenderer::Get()->GetRecordedCombinerStates() );

			pspDebugScreenPrintf( "Combiner States in use:\n" );
			pspDebugScreenPrintf( "   Use O to toggle on/off:\n" );

			u32		idx( 0 );
			u64		selected_state( 0 );
			for(std::set<u64>::const_iterator it = combiner_states.begin(); it != combiner_states.end(); ++it)
			{
				u64		state( *it );

				char	buf[200];

				bool	selected( idx == selected_idx );
				bool	disabled( PSPRenderer::Get()->IsCombinerStateDisabled( state ) );
				bool	unhandled( PSPRenderer::Get()->IsCombinerStateUnhandled( state ) );
				u32		text_col;

				if(selected)
				{
					selected_state = state;
					text_col = MAKE_COLOUR_U32( 255, 255, 0 );
				}
				else if(disabled)
				{
					text_col = MAKE_COLOUR_U32( 0, 255, 0 );
				}
				else if(unhandled)
				{
					text_col = MAKE_COLOUR_U32( 255, 0, 255 );
				}
				else
				{
					text_col = MAKE_COLOUR_U32( 255, 255, 255 );
				}

				pspDebugScreenSetTextColor( text_col );
				sprintf( buf, " %c%08x%08x\n", selected ? '*' : ' ', u32(state >> 32), u32(state) );

				pspDebugScreenPrintData( buf, strlen( buf ) );					
				//pspDebugScreenPrintf( "  %08x%08x\n", u32(state >> 32), u32(state) );

				idx++;
			}
			pspDebugScreenSetTextColor( MAKE_COLOUR_U32( 255, 255, 255 ) );

			u32		state_count( combiner_states.size() );

			if(oldButtons != pad.Buttons)
			{
				if(pad.Buttons & PSP_CTRL_UP)		{ selected_idx = (selected_idx > 0) ? selected_idx - 1 : selected_idx; }
				if(pad.Buttons & PSP_CTRL_DOWN)		{ selected_idx = (selected_idx < state_count-1) ? selected_idx + 1 : selected_idx; }			

				if(pad.Buttons & PSP_CTRL_CIRCLE)
				{
					if(selected_state != 0)
					{
						PSPRenderer::Get()->ToggleDisableCombinerState( selected_state );
					}
				}
			}
		}

		if(oldButtons != pad.Buttons)
		{
			if(pad.Buttons & PSP_CTRL_SELECT)
			{
				menu_button_pressed = true;
			}

			if(pad.Buttons & PSP_CTRL_CROSS)
			{
				show_combiner_states = !show_combiner_states;
			}
		}

		oldButtons = pad.Buttons;
	}

	PSPRenderer::Get()->SetRecordCombinerStates( false );
}
#endif // DAEDALUS_DEBUG_DISPLAYLIST

/******************************************************************

*******************************************************************/ 
void Graphics_ProcessDList(void)
{
	g_CritialSection.Lock();

	if (gDropAllTextures)
	{
		CTextureCache::Get()->DropTextures();
		gDropAllTextures = false;
	}

#ifdef DAEDALUS_DEBUG_DISPLAYLIST
	if(gDebugDisplayList)
	{
		Graphics_DebugDList();
		gDebugDisplayList = false;
	}
	else
#endif
	{
		DLParser_Process();
	}

	g_CritialSection.Unlock();
}

/******************************************************************
  Function: RomClosed
  Purpose:  This function is called when a rom is closed.
  input:    none
  output:   none
*******************************************************************/ 
void Graphics_RomClosed (void)
{
	g_CritialSection.Lock();

	if ( CTextureCache::IsAvailable() )
	{
		CTextureCache::Get()->DropTextures();
	}

	g_CritialSection.Unlock();
}

/******************************************************************
  Function: RomOpen
  Purpose:  This function is called when a rom is open. (from the 
            emulation thread)
  input:    none
  output:   none
*******************************************************************/ 

void Graphics_RomOpen (void)
{
	g_CritialSection.Lock();

	g_CritialSection.Unlock();
}


//*****************************************************************************
// Execute an arbitary command (e.g. "Take ScreenShot", "Dump DList")
// pszCommand: [in] Command name
// ppResult: [in/out] Optional auxillary data
// Returns standard HRESULT (S_OK, E_FAIL, E_OUTOFMEMORY etc)
//*****************************************************************************

HRESULT Graphics_ExecuteCommand( const CHAR * pszCommand, void ** ppResult )
{
	bool ok( false );

	g_CritialSection.Lock();

	if ( stricmp( pszCommand, DAEDALUS_GFX_SETDAEDALUSVERSION ) == 0 )
	{
		g_dwDaedalusVersion = (DWORD)*ppResult;
		ok = true;
	}
	else if ( stricmp( pszCommand, DAEDALUS_GFX_GETDAEDALUSPLUGINVERSION ) == 0 )
	{
		*ppResult = (void*)DAEDALUS_GFX_PLUGIN_VERSION;
		ok = true;
	}
	else if ( stricmp( pszCommand, DAEDALUS_GFX_DROPTEXTURES ) == 0 )
	{
		gDropAllTextures = true;
		ok = true;
	}
	else if ( stricmp( pszCommand, DAEDALUS_GFX_DUMPTEXTURES ) == 0 )
	{
		CTextureCache::Get()->SetDumpTextures( TRUE );
		gDropAllTextures = true;		// Drop textures so that we can re-create
		ok = true;
	}
	else if ( stricmp( pszCommand, DAEDALUS_GFX_NODUMPTEXTURES ) == 0 )
	{
		CTextureCache::Get()->SetDumpTextures( FALSE );
		ok = true;
	}
	else if ( stricmp( pszCommand, DAEDALUS_GFX_USENEWCOMBINER ) == 0)
	{
		g_bDisplayNewCombiner = (BOOL)*ppResult;
	}
	else if ( stricmp( pszCommand, DAEDALUS_GFX_DUMPDL ) == 0 )
	{
		DLParser_DumpNextDisplayList();
	}
	else if ( stricmp( pszCommand, DAEDALUS_GFX_DEBUGDL ) == 0 )
	{
		gDebugDisplayList = true;
	}
	g_CritialSection.Unlock();
	return ok ? 0 : -1;
}







/*
void IMainWindow::SizeForDisplay()
{
	RECT rcScreen;
	RECT rcStatus;

	// This might not be the best flag to switch on - ]
	// maybe we could check g_bD3DXReady?
	if (g_bCPURunning)
	{
		SetRect(&rcScreen, 0,0, g_dwDisplayWidth, g_dwDisplayHeight);

		// Calculate window size to give desired window size...
		AdjustWindowRectEx(&rcScreen, WS_OVERLAPPEDWINDOW, TRUE, 0);

		// Add on enough space for the status bar
		GetClientRect(m_hWndStatus, &rcStatus);
		rcScreen.bottom += (rcStatus.bottom - rcStatus.top);

		SetWindowPos(m_hWnd, 0,
			rcScreen.left, rcScreen.top,
			rcScreen.right-rcScreen.left, rcScreen.bottom-rcScreen.top,
			SWP_NOMOVE|SWP_NOACTIVATE|SWP_NOZORDER);
	}
	else
	{
		rcScreen = g_rcMainWindow;

		SetWindowPos(m_hWnd, 0,
			rcScreen.left, rcScreen.top,
			rcScreen.right-rcScreen.left, rcScreen.bottom-rcScreen.top,
			SWP_NOACTIVATE|SWP_NOZORDER);
	}
}
*/
