/*

  Copyright (C) 2001 StrmnNrmn

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"
#include "GraphicsContext.h"
#include "PngUtil.h"
#include "Debug\DBGConsole.h"
#include "Debug\Debug.h"
#include "Debug\Dump.h"

#include "TextureCache.h"
#include "VideoMemoryManager.h"

#include <pspgu.h>
#include <pspdisplay.h>
#include <pspdebug.h>
#include <png.h>

#ifndef SAFE_RELEASE
#define SAFE_RELEASE(p)  { if(p) { (p)->Release();     (p)=NULL; } }
#endif

namespace
{
	const char *	gScreenDumpRootPath = "screen_dumps";
	const char *	gScreenDumpDumpPathFormat = "sd%04d.png";
}


unsigned int __attribute__((aligned(16))) list[262144];

#define BUF_WIDTH (512)
#define SCR_WIDTH (480)
#define SCR_HEIGHT (272)
#define PIXEL_SIZE (4) /* change this if you change to another screenmode */
#define FRAME_SIZE (BUF_WIDTH * SCR_HEIGHT * PIXEL_SIZE)
#define DEPTH_SIZE (BUF_WIDTH * SCR_HEIGHT * 2)


// Implementation
class IGraphicsContext : public CGraphicsContext
{
public:
	virtual ~IGraphicsContext();
	
	BOOL Ready() { return m_bReady; }
	
	void SetReady(BOOL bReady) { m_bReady = bReady; }
	
	void Clear(bool clear_screen, bool clear_depth);
	void Clear(u32 frame_buffer_col, u32 depth);

	void BeginFrame();
	void EndFrame();
	bool UpdateFrame( bool wait_for_vbl );
	bool GetBufferSize(DWORD * pdwWidth, DWORD * pdwHeight);
	bool Initialise();
		
	void DumpScreenShot();
	
private:
	static void			SaveScreenshot( const char* filename );
	
protected:
	friend class CSingleton< CGraphicsContext >;

	IGraphicsContext();

	BOOL				m_bReady;
	BOOL				m_bActive;	
};

//*************************************************************************************
// 
//*************************************************************************************
template<> bool CSingleton< CGraphicsContext >::Create()
{	
	IGraphicsContext * pNewInstance;
	
	DAEDALUS_ASSERT_Q(mpInstance == NULL);
	
	pNewInstance = new IGraphicsContext();
	if (!pNewInstance)
	{
		return false;
	}

	pNewInstance->Initialise();
	
	mpInstance = pNewInstance;
	
	return true;
};


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

IGraphicsContext::IGraphicsContext() :
	m_bReady(FALSE), 
	m_bActive(FALSE)
{
}

//*****************************************************************************
//
//*****************************************************************************
IGraphicsContext::~IGraphicsContext()
{
	sceGuTerm();
}


//*****************************************************************************
//
//*****************************************************************************
void IGraphicsContext::Clear(bool clear_screen, bool clear_depth)
{
	int	flags( 0 );

	if(clear_screen)
		flags |= GU_COLOR_BUFFER_BIT;
	if(clear_depth)
		flags |= GU_DEPTH_BUFFER_BIT;

	sceGuClearColor(0xff000000);
	sceGuClearDepth(0);				// 1?
	sceGuClear(flags);
}

//*****************************************************************************
//
//*****************************************************************************
void IGraphicsContext::Clear(u32 frame_buffer_col, u32 depth)
{
	sceGuClearColor(frame_buffer_col);
	sceGuClearDepth(depth);				// 1?
	sceGuClear(GU_COLOR_BUFFER_BIT|GU_DEPTH_BUFFER_BIT);
}

//*****************************************************************************
//
//*****************************************************************************
void IGraphicsContext::BeginFrame()
{
	sceGuStart(GU_DIRECT,list);
}

//*****************************************************************************
//
//*****************************************************************************
void IGraphicsContext::EndFrame()
{
	sceGuFinish();
}

//*****************************************************************************
//
//*****************************************************************************
bool IGraphicsContext::UpdateFrame( bool wait_for_vbl )
{
	sceGuSync(0,0);
	if(wait_for_vbl)
	{
		sceDisplayWaitVblankStart();
	}
	sceGuSwapBuffers();

	return true;
}


//*****************************************************************************
// Save current visible screen as PNG
// From Shazz/71M - thanks guys!
//*****************************************************************************
void IGraphicsContext::SaveScreenshot( const char* filename )
{
	u32* vram32;
	u16* vram16;
	int bufferwidth;
	int pixelformat;
	int unknown;
	int i, x, y;
	png_structp png_ptr;
	png_infop info_ptr;
	FILE* fp;
	u8* line;

	fp = fopen(filename, "wb");
	if (!fp) return;
	png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if (!png_ptr) return;
	info_ptr = png_create_info_struct(png_ptr);
	if (!info_ptr) {
		png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
		fclose(fp);
		return;
	}
	png_init_io(png_ptr, fp);
	png_set_IHDR(png_ptr, info_ptr, SCR_WIDTH, SCR_HEIGHT,
		8, PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE,
		PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
	png_write_info(png_ptr, info_ptr);
	line = (u8*) malloc(SCR_WIDTH * 3);
	sceDisplayWaitVblankStart();  // if framebuf was set with PSP_DISPLAY_SETBUF_NEXTFRAME, wait until it is changed
	sceDisplayGetFrameBuf((void**)&vram32, &bufferwidth, &pixelformat, &unknown);
	vram16 = (u16*) vram32;
	for (y = 0; y < SCR_HEIGHT; y++) {
		for (i = 0, x = 0; x < SCR_WIDTH; x++) {
			u32 color = 0;
			u8 r = 0, g = 0, b = 0;
			switch (pixelformat) {
				case PSP_DISPLAY_PIXEL_FORMAT_565:
					color = vram16[x + y * bufferwidth];
					r = (color & 0x1f) << 3;
					g = ((color >> 5) & 0x3f) << 2 ;
					b = ((color >> 11) & 0x1f) << 3 ;
					break;
				case PSP_DISPLAY_PIXEL_FORMAT_5551:
					color = vram16[x + y * bufferwidth];
					r = (color & 0x1f) << 3;
					g = ((color >> 5) & 0x1f) << 3 ;
					b = ((color >> 10) & 0x1f) << 3 ;
					break;
				case PSP_DISPLAY_PIXEL_FORMAT_4444:
					color = vram16[x + y * bufferwidth];
					r = (color & 0xf) << 4;
					g = ((color >> 4) & 0xf) << 4 ;
					b = ((color >> 8) & 0xf) << 4 ;
					break;
				case PSP_DISPLAY_PIXEL_FORMAT_8888:
					color = vram32[x + y * bufferwidth];
					r = color & 0xff;
					g = (color >> 8) & 0xff;
					b = (color >> 16) & 0xff;
					break;
			}
			line[i++] = r;
			line[i++] = g;
			line[i++] = b;
		}
		png_write_row(png_ptr, line);
	}
	free(line);
	png_write_end(png_ptr, info_ptr);
	png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
	fclose(fp);
}

//*****************************************************************************
//
//*****************************************************************************
void IGraphicsContext::DumpScreenShot()
{
	char szFilePath[MAX_PATH+1];
	char szDumpDir[MAX_PATH+1];

#ifdef DAEDALUS_PSP
	strcpy(szDumpDir, gScreenDumpRootPath);			// Some issues with subdirs :(
#else
	PathCombine(szDumpDir, gScreenDumpRootPath, g_ROM.settings.szGameName);
#endif

	Dump_GetDumpDirectory(szFilePath, szDumpDir);

	char	unique_filename[MAX_PATH+1];
	u32 count = 0;
	do
	{
		char	test_name[MAX_PATH+1];

		sprintf(test_name, "sd%04d.png", count++);

		strcpy( unique_filename, szFilePath );
		PathAppend( unique_filename, test_name );

	} while( PathFileExists( unique_filename ) );

	SaveScreenshot( unique_filename );

}


//*****************************************************************************
//
//*****************************************************************************
bool IGraphicsContext::GetBufferSize(DWORD * pdwWidth, DWORD * pdwHeight)
{
	*pdwWidth = SCR_WIDTH;
	*pdwHeight = SCR_HEIGHT;

	return true;
}
//*****************************************************************************
//
//*****************************************************************************
bool IGraphicsContext::Initialise()
{
	sceGuInit();

	u8 *	p_draw_buffer( (u8*)CVideoMemoryManager::Get()->Alloc( FRAME_SIZE ) );
	u8 *	p_disp_buffer( (u8*)CVideoMemoryManager::Get()->Alloc( FRAME_SIZE ) );
	u8 *	p_depth_buffer( (u8*)CVideoMemoryManager::Get()->Alloc( DEPTH_SIZE ) );

	pspDebugScreenPrintf( "Allocated %d bytes of memory for draw buffer at %p\n", FRAME_SIZE, p_draw_buffer );
	pspDebugScreenPrintf( "Allocated %d bytes of memory for draw buffer at %p\n", FRAME_SIZE, p_disp_buffer );
	pspDebugScreenPrintf( "Allocated %d bytes of memory for draw buffer at %p\n", DEPTH_SIZE, p_depth_buffer );

	sceGuStart(GU_DIRECT,list);
	sceGuDrawBuffer(GU_PSM_8888,(u8*)(p_draw_buffer - (u8*)sceGeEdramGetAddr()),BUF_WIDTH);
	sceGuDispBuffer(SCR_WIDTH,SCR_HEIGHT,(u8*)(p_disp_buffer - (u8*)sceGeEdramGetAddr()),BUF_WIDTH);
	sceGuDepthBuffer((u8*)(p_depth_buffer - (u8*)sceGeEdramGetAddr()),BUF_WIDTH);
	sceGuOffset(2048 - (SCR_WIDTH/2),2048 - (SCR_HEIGHT/2));
	sceGuViewport(2048,2048,SCR_WIDTH,SCR_HEIGHT);
	sceGuDepthRange(0xc350,0x2710);
	sceGuScissor(0,0,SCR_WIDTH,SCR_HEIGHT);
	sceGuEnable(GU_SCISSOR_TEST);
	sceGuDepthFunc(GU_GEQUAL);
	sceGuEnable(GU_DEPTH_TEST);
	sceGuFrontFace(GU_CW);
	sceGuShadeModel(GU_SMOOTH);
	sceGuEnable(GU_CULL_FACE);
	sceGuEnable(GU_TEXTURE_2D);
	sceGuEnable(GU_CLIP_PLANES);
	sceGuFinish();
	sceGuSync(0,0);

	sceDisplayWaitVblankStart();
	sceGuDisplay(GU_TRUE);

	// The app is ready to go
	m_bReady = TRUE;
    m_bActive = TRUE;
    return true;
}


