/*
Copyright (C) 2005 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"
#include "NativeTexture.h"

#include "VideoMemoryManager.h"
#include "DaedMathUtil.h"

#include <pspgu.h>
#include <pspdebug.h>
#include <pspkernel.h>

void swizzle_fast(u8* out, const u8* in, unsigned int width, unsigned int height)
{
	unsigned int blockx, blocky;
	unsigned int j;

	unsigned int width_blocks = (width / 16);
	unsigned int height_blocks = (height / 8);

	unsigned int src_pitch = (width-16)/4;
	unsigned int src_row = width * 8;

	const u8* ysrc = in;
	u32* dst = (u32*)out;

	for (blocky = 0; blocky < height_blocks; ++blocky)
	{
		const u8* xsrc = ysrc;
		for (blockx = 0; blockx < width_blocks; ++blockx)
		{
			const u32* src = (u32*)xsrc;
			for (j = 0; j < 8; ++j)
			{
				*(dst++) = *(src++);
				*(dst++) = *(src++);
				*(dst++) = *(src++);
				*(dst++) = *(src++);
				src += src_pitch;
			}
			xsrc += 16;
		}
		ysrc += src_row;
	}
}

//
//	Swizzle the texture, explicitly setting the texture colour to the specified value (alpha is left alone!)
//
void swizzle_fast_setcolour(u8* out, const u8* in, unsigned int width, unsigned int height, u32 colour)
{
	unsigned int blockx, blocky;
	unsigned int j;

	unsigned int width_blocks = (width / 16);
	unsigned int height_blocks = (height / 8);

	unsigned int src_pitch = (width-16)/4;
	unsigned int src_row = width * 8;

	const u8* ysrc = in;
	u32* dst = (u32*)out;

	static const u32	COLOUR_MASK = 0x00ffffff;
	static const u32	ALPHA_MASK  = 0xff000000;

	// For speed, clear out the alpha value
	u32 just_colour( colour & COLOUR_MASK );

	for (blocky = 0; blocky < height_blocks; ++blocky)
	{
		const u8* xsrc = ysrc;
		for (blockx = 0; blockx < width_blocks; ++blockx)
		{
			const u32* src = (u32*)xsrc;
			for (j = 0; j < 8; ++j)
			{
				*(dst++) = (*(src++) & ALPHA_MASK) | just_colour;
				*(dst++) = (*(src++) & ALPHA_MASK) | just_colour;
				*(dst++) = (*(src++) & ALPHA_MASK) | just_colour;
				*(dst++) = (*(src++) & ALPHA_MASK) | just_colour;
				src += src_pitch;
			}
			xsrc += 16;
		}
		ysrc += src_row;
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CNativeTexture::CorrectDimensions( u32 width, u32 height, u32 & adjusted_width, u32 & adjusted_height )
{
	static const u32 MIN_TEXTURE_WIDTH = 1;
	static const u32 MIN_TEXTURE_HEIGHT = 1;

	adjusted_width = Max( GetNextPowerOf2( width ), MIN_TEXTURE_WIDTH );
	adjusted_height = Max( GetNextPowerOf2( height ), MIN_TEXTURE_HEIGHT );
}

//*****************************************************************************
//
//*****************************************************************************
CNativeTexture *	CNativeTexture::CreateFromRaw( const char * p_filename, u32 width, u32 height )
{
	u32		adjusted_width;
	u32		adjusted_height;

	CorrectDimensions( width, height, adjusted_width, adjusted_height );

	//
	// Create a texture (do this first to help minimise memory fragmentation)
	//
	CNativeTexture *	p_texture( new CNativeTexture( adjusted_width, adjusted_height, 4 ) );

	//
	//	Read in the data (taking care that it's width/height is different from what we expect)
	//
	FILE * fh( fopen( p_filename, "rb" ) );
	if(fh != NULL)
	{
		u32 *	p_data( new u32[ adjusted_width * adjusted_height ] );
		for(u32 y = 0; y < height; ++y)
		{
			u32 *	p_row( &p_data[ y * adjusted_width ] );

			fread( p_row, width, sizeof(u32), fh );

			// Rest of the row is garbage
		}

		fclose( fh );
		p_texture->SetData( p_data, adjusted_width, adjusted_height );

		delete [] p_data;

	}
	else
	{
		p_texture->Release();
		p_texture = NULL;
	}
	
	return p_texture;
}

//*****************************************************************************
//
//*****************************************************************************
CNativeTexture *	CNativeTexture::CreateFromAlpha( const char * p_filename, u32 width, u32 height )
{
	CNativeTexture *	p_texture( new CNativeTexture( width, height, 4 ) );
	if(p_texture != NULL)
	{
		FILE * fh( fopen( p_filename, "rb" ) );
		if(fh != NULL)
		{
			u32 *	p_data( p_texture->mpData );
			u8 *	p_row( new u8[ width ] );

			for(u32 y = 0; y < height; ++y)
			{
				fread( p_row, width, 1, fh );
				for(u32 x = 0; x < width; ++x)
				{
					*(p_data++) = ((u32)p_row[x] << 24) | 0x00ffffff;
				}
			}

			delete [] p_row;
			fclose( fh );
		}

		p_texture->mIsSwizzled = false;
	}

	return p_texture;
}


//*****************************************************************************
//
//*****************************************************************************
CNativeTexture::CNativeTexture( u32 w, u32 h, u8 channels ) : mWidth( w ), mHeight( h ), mIsSwizzled( true ), mPixelSize( channels )
{
	u32 * pdata( (u32*)CVideoMemoryManager::Get()->Alloc( mWidth * mHeight * sizeof(u32) ) );
	if(pdata != NULL)
	{
		mpData = pdata;
		mIsVidMem = true;
	}
	else
	{
		mpData = new u32[ mWidth * mHeight ];
		mIsVidMem = false;
	}
}

//*****************************************************************************
//
//*****************************************************************************
CNativeTexture::CNativeTexture( u32 w, u32 h ) : mWidth( w ), mHeight( h ), mIsSwizzled( true ), mPixelSize( 4 )
{
	u32 * pdata( (u32*)CVideoMemoryManager::Get()->Alloc( mWidth * mHeight * mPixelSize ) );
	if(pdata != NULL)
	{
		mpData = pdata;
		mIsVidMem = true;
	}
	else
	{
		mpData = new u32[ mWidth * mHeight ];
		mIsVidMem = false;
	}

}

//*****************************************************************************
//
//*****************************************************************************
CNativeTexture::~CNativeTexture()
{
	if(mIsVidMem)
	{
		CVideoMemoryManager::Get()->Free( mpData );
	}
	else
	{
		delete [] mpData;
	}
}

//*****************************************************************************
//
//*****************************************************************************
void	CNativeTexture::InstallTexture()
{
	sceGuEnable(GU_TEXTURE_2D);
	sceGuTexMode(GU_PSM_8888,0,0,mIsSwizzled ? 1 : 0);		// maxmips/a2/swizzle = 0
	sceGuTexImage(0,mWidth,mHeight,mWidth,mpData);
}

//*****************************************************************************
//
//*****************************************************************************
void	CNativeTexture::SetData( void * p_data, u32 w, u32 h )
{
	DAEDALUS_ASSERT_Q( w == mWidth );
	DAEDALUS_ASSERT_Q( h == mHeight );

	if(mHeight >= 8)
	{
		swizzle_fast( reinterpret_cast< u8 * >( mpData ), reinterpret_cast< u8 * >( p_data ), mWidth*mPixelSize, mHeight );
		mIsSwizzled = true;
	}
	else
	{
		memcpy( mpData, p_data, w*h*mPixelSize );
		mIsSwizzled = false;
	}

	//sceKernelDcacheWritebackAll();
	//sceGuTexFlush();
}

//*****************************************************************************
//
//*****************************************************************************
void	CNativeTexture::SetDataForceColour( void * p_data, u32 w, u32 h, u32 colour )
{
	DAEDALUS_ASSERT_Q( w == mWidth );
	DAEDALUS_ASSERT_Q( h == mHeight );

	swizzle_fast_setcolour( reinterpret_cast< u8 * >( mpData ), reinterpret_cast< u8 * >( p_data ), mWidth*mPixelSize, mHeight, colour );
	mIsSwizzled = true;

	//sceKernelDcacheWritebackAll();
	//sceGuTexFlush();
}

//*****************************************************************************
//
//*****************************************************************************
u32	CNativeTexture::GetVideoMemoryUsage() const
{
	if( mIsVidMem )
	{
		return mWidth * mHeight * mPixelSize;
	}

	return 0;
}

//*****************************************************************************
//
//*****************************************************************************
u32	CNativeTexture::GetSystemMemoryUsage() const
{
	if( !mIsVidMem )
	{
		return mWidth * mHeight * mPixelSize;
	}

	return 0;
}
