#include "stdafx.h"
#include "PolyTesselator.h"
#include "DaedalusVtx.h"

#include "Matrix4x4.h"

#include "DebugDisplayList.h"

//*****************************************************************************
//
//*****************************************************************************
static const u32	MAX_TESSELATION_DEPTH( 4 );
static const float	MAX_SIZE_ACCEPTED( 1.0f );

#define CC_NEG_Y		(1<<0)
#define CC_POS_Y		(1<<1)
#define CC_NEG_X		(1<<2)
#define CC_POS_X		(1<<3)

//*****************************************************************************
//
//*****************************************************************************
inline float min3( float a, float b, float c )
{
	if(a<=b && a<=c)
	{
		return a;
	}
	if(b<=c)
	{
		return b;
	}
	return c;
}

//*****************************************************************************
//
//*****************************************************************************
inline float max3( float a, float b, float c )
{
	if(a>=b && a>=c)
	{
		return a;
	}
	if(b>=c)
	{
		return b;
	}
	return c;
}

//*****************************************************************************
//
//*****************************************************************************
static u32	GetClipCode( const v3 & p )
{
	u32		code( 0 );

	if(p.x < -1) code |= CC_NEG_X;
	if(p.x > +1) code |= CC_POS_X;
	if(p.y < -1) code |= CC_NEG_Y;
	if(p.y > +1) code |= CC_POS_Y;

	return code;
}

//*****************************************************************************
//
//*****************************************************************************
static u32 TesselateTriangle( u32 depth, DaedalusVtx * p_out, u32 out_idx, const DaedalusVtx & a, const DaedalusVtx & b, const DaedalusVtx & c, const Matrix4x4 & projection )
{
	bool		accepted( false );
	bool		rejected( false );

	//if(depth > 0)
	{
		if(g_hRDPDumpHandle != NULL)
		{
			DL_PF( " %d: Recursing - count is %d (subdividing (%.2f,%.2f,%.2f) (%.2f,%.2f,%.2f) (%.2f,%.2f,%.2f)", depth, out_idx,
				a.pos.x, a.pos.y, a.pos.z, b.pos.x, b.pos.y, b.pos.z, c.pos.x, c.pos.y, c.pos.z );
		}
	}

	//
	//	If we've hit the maximum recursion depth, just accept as it is
	//
	if(depth >= MAX_TESSELATION_DEPTH)
	{
		accepted = true;
	}
	else
	{
		//
		//	Project the vertices and figure out if this triangle is fully onscreen
		//
		v3	projected_a( projection.Transform( a.pos ) );			// Homegeneous Transform - divides out w
		v3	projected_b( projection.Transform( b.pos ) );
		v3	projected_c( projection.Transform( c.pos ) );

		u32		cc_a( GetClipCode( projected_a ) );
		u32		cc_b( GetClipCode( projected_b ) );
		u32		cc_c( GetClipCode( projected_c ) );

		//
		//	Check the clip codes
		//
		if((cc_a | cc_b | cc_c) == 0)
		{
			//
			//	All inside
			//
			accepted = true;
		}
		else if((cc_a & cc_b & cc_c) != 0)
		{
			//
			//	All on same side of plane (we can reject this)
			//
			rejected = true;
		}
		else
		{
			v2	min_extent( min3( projected_a.x, projected_b.x, projected_c.x ), 
							min3( projected_a.y, projected_b.y, projected_c.y ) );
			v2	max_extent( max3( projected_a.x, projected_b.x, projected_c.x ), 
							max3( projected_a.y, projected_b.y, projected_c.y ) );

			v2	extent( max_extent - min_extent );

			//
			//	If the triangle is small enough then accept it
			//
			if(extent.x < MAX_SIZE_ACCEPTED && extent.y < MAX_SIZE_ACCEPTED)
			{
				accepted = true;
			}
		}
	}

	if(rejected)
	{
		//
		//	Do nothing
		//
	}
	else if(accepted)
	{
		//
		//	Immediately accept and move on
		//
		if(p_out != NULL)
		{
			p_out[out_idx++] = a;
			p_out[out_idx++] = b;
			p_out[out_idx++] = c;
		}
		else
		{
			out_idx += 3;
		}
	}
	else
	{
		//
		//	This triangle wasn't entirely on screen. Tesselate and recurse
		//
		DaedalusVtx	ab( a.Interpolate( b, 0.5f ) );
		DaedalusVtx	bc( b.Interpolate( c, 0.5f ) );
		DaedalusVtx	ca( c.Interpolate( a, 0.5f ) );

		out_idx = TesselateTriangle( depth + 1, p_out, out_idx,  a, ab, ca, projection );
		out_idx = TesselateTriangle( depth + 1, p_out, out_idx,  b, bc, ab, projection );
		out_idx = TesselateTriangle( depth + 1, p_out, out_idx,  c, ca, bc, projection );
		out_idx = TesselateTriangle( depth + 1, p_out, out_idx, ab, bc, ca, projection );
	}

	return out_idx;
}

//*****************************************************************************
//
//*****************************************************************************
u32	TesselateVerts( DaedalusVtx * p_out, const DaedalusVtx * p_in, const u16 * p_indices, u32 num_indices, const Matrix4x4 & projection )
{
	u32	out_idx( 0 );
	for(u32 t = 0; t+2 < num_indices; t += 3)
	{
		u16		idx_a( p_indices[t+0] );
		u16		idx_b( p_indices[t+1] );
		u16		idx_c( p_indices[t+2] );

		const DaedalusVtx &	a( p_in[idx_a] );
		const DaedalusVtx &	b( p_in[idx_b] );
		const DaedalusVtx &	c( p_in[idx_c] );

		out_idx = TesselateTriangle( 0, p_out, out_idx, a, b, c, projection );
	}

	return out_idx;
}
