/*
Copyright (C) 2001 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "stdafx.h"

#include "Texture.h"
#include "NativeTexture.h"
#include "PngUtil.h"
#include "DaedMathUtil.h"

#include "Debug\DBGConsole.h"

#include <pspgu.h>
#include <psputils.h>

//*****************************************************************************
//
//*****************************************************************************
class ITexture : public CTexture
{
	public:
		ITexture();
		virtual ~ITexture();

		bool					Initialise(DWORD dwWidth, DWORD dwHeight);
		CNativeTexture *		GetTexture()					{ return m_pTexture; }
		CNativeTexture *		GetRecolouredTexture( u32 colour );

#ifdef LIBPNG_SUPPORT
		BOOL					DumpImageAsPNG(LPCTSTR szFileName, DWORD dwWidth, DWORD dwHeight);
#endif
		DWORD					GetPixelSize();

		virtual	u32				GetWidth() const				{ return m_dwCorrectedWidth; }
		virtual	u32				GetHeight() const				{ return m_dwCorrectedHeight; }

		virtual u32				GetVideoMemoryUsage() const;
		virtual u32				GetSystemMemoryUsage() const;

		// Provides access to "surface"
		BOOL					StartUpdate(DrawInfo *di);
		void					EndUpdate(DrawInfo *di);

	protected:

		u8 *					mpTexels;
		CNativeTexture *		m_pTexture;
		CNativeTexture *		mpRecolouredTexture;
		
		DWORD					m_dwWidth;			// The requested Texture w/h
		DWORD					m_dwHeight;

		u32						m_dwCorrectedWidth;	// What was actually created
		u32						m_dwCorrectedHeight;
};



//*****************************************************************************
//
//*****************************************************************************
CTexture * CTexture::Create( DWORD dwWidth, DWORD dwHeight )
{
	ITexture * pTexture = new ITexture();

	if (!pTexture->Initialise(dwWidth, dwHeight))
	{
		delete pTexture;
		return NULL;
	}

	return pTexture;
}

//*****************************************************************************
//
//*****************************************************************************
ITexture::ITexture() :
	mpTexels(NULL),
	m_pTexture(NULL),
	mpRecolouredTexture(NULL),
	m_dwWidth(0),
	m_dwHeight(0),
	m_dwCorrectedWidth(0),
	m_dwCorrectedHeight(0)
{
}

//*****************************************************************************
//
//*****************************************************************************
ITexture::~ITexture(void)
{
	delete [] mpTexels;

	if (m_pTexture)
	{
		m_pTexture->Release();
	}
	if(mpRecolouredTexture)
	{
		mpRecolouredTexture->Release();
	}
}

//*****************************************************************************
//
//*****************************************************************************
bool ITexture::Initialise(DWORD dwWidth, DWORD dwHeight)
{
	DAEDALUS_ASSERT_Q(m_pTexture == NULL);

	CNativeTexture::CorrectDimensions( dwWidth, dwHeight, m_dwCorrectedWidth, m_dwCorrectedHeight );

	m_pTexture = new CNativeTexture( m_dwCorrectedWidth, m_dwCorrectedHeight );

	if(m_pTexture != NULL)
	{
		m_dwWidth = dwWidth;
		m_dwHeight = dwHeight;
		return true;
	}

	return false;
}

//*****************************************************************************
//
//*****************************************************************************
DWORD ITexture::GetPixelSize()
{
	return 4;
}
//*****************************************************************************
//
//*****************************************************************************
CNativeTexture *	ITexture::GetRecolouredTexture( u32 colour )
{
	if(mpRecolouredTexture == NULL)
	{
		mpRecolouredTexture = new CNativeTexture( m_dwCorrectedWidth, m_dwCorrectedHeight );

		if(mpRecolouredTexture != NULL)
		{
			mpRecolouredTexture->SetDataForceColour( mpTexels, m_dwCorrectedWidth, m_dwCorrectedHeight, colour );
		}
	}

	return mpRecolouredTexture;
}

//*****************************************************************************
// Lock the texture so we can start updating it
//*****************************************************************************
BOOL ITexture::StartUpdate(DrawInfo *di)
{
	if (m_pTexture != NULL)
	{
		u32		buffer_size( m_dwCorrectedWidth * m_dwCorrectedHeight * GetPixelSize() );

		mpTexels = new u8[ buffer_size ];

		if(mpTexels != NULL)
		{
			// Return a temporary buffer to use
			di->lpSurface = mpTexels;
			di->dwWidth = m_dwCorrectedWidth;
			di->dwHeight = m_dwCorrectedHeight;
			di->lPitch = m_dwCorrectedWidth * GetPixelSize();

			return true;
		}
	}

	return false;
}

//*****************************************************************************
// Finish updating the texture and copy the memory to the native surface
//*****************************************************************************
void ITexture::EndUpdate(DrawInfo *di)
{
	if (m_pTexture != NULL)
	{
		m_pTexture->SetData( mpTexels, m_dwCorrectedWidth, m_dwCorrectedHeight );
	}
}

//*****************************************************************************
//
//*****************************************************************************
u32	ITexture::GetVideoMemoryUsage() const
{
	if(m_pTexture != NULL)
	{
		return m_pTexture->GetVideoMemoryUsage();
	}

	return 0;
}

//*****************************************************************************
//
//*****************************************************************************
u32	ITexture::GetSystemMemoryUsage() const
{
	if(m_pTexture != NULL)
	{
		return m_pTexture->GetSystemMemoryUsage();
	}

	return 0;
}

#ifdef LIBPNG_SUPPORT
//*****************************************************************************
//
//*****************************************************************************
BOOL ITexture::DumpImageAsPNG(LPCTSTR szFileName, DWORD dwWidth, DWORD dwHeight)
{
	DWORD * pB = new DWORD [dwWidth * dwHeight];
	if (pB == NULL)
	{
		return FALSE;
	}

	DrawInfo di;
	
	if (!StartUpdate(&di))
	{
		delete [] pB;
		return FALSE;
	}
	
	DWORD *pSrc = (DWORD *)di.lpSurface;
	LONG lPitch = di.lPitch;

	DWORD index = 0;

	for (DWORD y = 0; y < dwHeight; y++)
	{
		DWORD * dwSrc = (DWORD *)((BYTE *)pSrc + y*lPitch);
		for (DWORD x = 0; x < dwWidth; x++)
		{
			pB[index++] = dwSrc[x];
		}
	}
	EndUpdate(&di);
	
	PngSaveImage(szFileName, (png_byte *)pB, dwWidth, dwHeight);

	delete [] pB;

	return TRUE;
}

#endif // LIBPNG_SUPPORT
