/*
Copyright (C) 2001 StrmnNrmn

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
	  
*/

// Manages textures for RDP code
// Uses a HashTable (hashing on TImg) to allow quick access
//  to previously used textures

#include "stdafx.h"

//#include "memory.h"
#include "TextureCache.h"
#include "ConvertImage.h"
#include "RDP.h"

#include "ultra_gbi.h"		// For image size/formats
#include "Utility\Profiler.h"
#include "Core\ROM.h"
#include "Core\Memory.h"

#include "Debug\DaedalusAssert.h"

#include "DaedHash.h"

#include <pspdebug.h>

extern u32 gRDPFrame;

// Returns the first prime greater than or equal to nFirst
static s32 GetNextPrime(s32 nFirst)
{
	s32 nCurrent;
	
	s32 i;
	
	nCurrent = nFirst;
	
	// Just make sure it's odd
	if ((nCurrent % 2) == 0)
		nCurrent++;
	
	for (;;)
	{
		s32 nSqrtCurrent;
		BOOL bIsComposite;
		
		// nSqrtCurrent = nCurrent^0.5 + 1 (round up)
		nSqrtCurrent = (s32)sqrt((double)nCurrent) + 1;
		
		bIsComposite = FALSE;
		
		// Test all odd numbers from 3..nSqrtCurrent
		for (i = 3; i <= nSqrtCurrent; i+=2)
		{
			if ((nCurrent % i) == 0)
			{
				bIsComposite = TRUE;
				break;
			}
		}
		
		if (!bIsComposite)
		{			
			return nCurrent;
		}
		
		// Select next odd candidate...
		nCurrent += 2;
	}
	
}


// Implementation

typedef struct TextureEntry
{
	TextureEntry() : pTexture(NULL), pMirroredTexture(NULL) {}

	~TextureEntry()
	{
		if (pTexture)
		{
			pTexture->Release();
			pTexture = NULL;
		}
		if (pMirroredTexture)
		{
			pMirroredTexture->Release();
			pMirroredTexture = NULL;
		}
	}

	u32		GetVideoMemoryUsage() const
	{
		u32	usage( 0 );
		if( pTexture != NULL )
		{
			usage += pTexture->GetVideoMemoryUsage();
		}
		if( pMirroredTexture != NULL )
		{
			usage += pMirroredTexture->GetVideoMemoryUsage();
		}
		return usage;
	}

	u32		GetSystemMemoryUsage() const
	{
		u32	usage( 0 );
		if( pTexture != NULL )
		{
			usage += pTexture->GetSystemMemoryUsage();
		}
		if( pMirroredTexture != NULL )
		{
			usage += pMirroredTexture->GetSystemMemoryUsage();
		}
		return usage;
	}


	//
	// Data
	//
	struct TextureEntry * pNext;

	TextureInfo ti;
	DWORD dwCRC;

	DWORD dwUses;			// Total times used (for stats)
	DWORD FrameLastUsed;	// Frame # that this was last used

	CTexture	* pTexture;
	// TODO: when nXTimes and nYTimes in the mirror emulator are correctly determined, these should be put here
	CTexture * pMirroredTexture;

} TextureEntry;



class ITextureCache : public CTextureCache
{
public:
	ITextureCache();
	~ITextureCache();
	
			bool			Initialise(u32 size);

			void			PurgeOldTextures();
			u32				CountOldTextures() const;
			void			DropTextures();

			void			SetDumpTextures( BOOL bDumpTextures ) { m_bDumpTextures = bDumpTextures; }
			BOOL			GetDumpTextures( ) const { return m_bDumpTextures; }


			bool			InvalidateDeviceObjects();
	
			CTexture *		GetTexture(TextureInfo * pti);
	
	virtual	void			DisplayStats();

protected:
			TextureEntry *	CreateEntry(TextureInfo * pti);
			void			AddTextureEntry(TextureEntry *pEntry);
			void			RemoveTextureEntry(TextureEntry * pEntry);
			
			void			AddToRecycleList(TextureEntry *pEntry);
			TextureEntry *	ReviveUsedTexture(DWORD dwWidth, DWORD dwHeight);
			
			
			TextureEntry *	GetEntry(TextureInfo * pti);
			TextureEntry *	GetEntryAndUpdate(TextureInfo * pti);
			
			DWORD			Hash(DWORD dwValue);
			
			void			DecompressTexture(TextureEntry * pEntry);
			void			DecompressTmemTexture(TextureEntry * pEntry);

			DWORD			CalculateCRC(TextureInfo * pti) const;

			void			GetUsedTextureStats( u32 * p_num_textures, u32 * p_video_memory_used, u32 * p_system_memory_used ) const;
			void			GetOldTextureStats( u32 * p_num_textures, u32 * p_video_memory_used, u32 * p_system_memory_used ) const;

protected:
	TextureEntry *			m_pFirstUsedSurface;
	TextureEntry **			m_pTextureHash;
	DWORD					m_dwTextureHashSize;
	BOOL					m_bDumpTextures;
	
};


// Interface

//*****************************************************************************
//
//*****************************************************************************

template<> bool CSingleton< CTextureCache >::Create()
{
	ITextureCache * pNewInstance;
	
	DAEDALUS_ASSERT_Q(mpInstance == NULL);
	
	pNewInstance = new ITextureCache();
	if (!pNewInstance)
	{
		return false;
	}
	else
	{
		if (!pNewInstance->Initialise(8000))
		{
			delete pNewInstance;
			pNewInstance = NULL;
		}
	}
	
	mpInstance = pNewInstance;
	return true;
}

//*************************************************************************************
//
//*************************************************************************************
ITextureCache::ITextureCache() :
	m_pFirstUsedSurface(NULL),
	m_pTextureHash(NULL),
	m_dwTextureHashSize(0),
	m_bDumpTextures(FALSE)
{
}

//*************************************************************************************
//
//*************************************************************************************
ITextureCache::~ITextureCache()
{
	DropTextures();
	
	while (m_pFirstUsedSurface)
	{
		TextureEntry * pVictim = m_pFirstUsedSurface;
		m_pFirstUsedSurface = pVictim->pNext;
		
		delete pVictim;
	}

	delete []m_pTextureHash;
	m_pTextureHash = NULL;	
}

//*************************************************************************************
// Called when the device-dependent objects are about to be lost.
//*************************************************************************************
bool ITextureCache::InvalidateDeviceObjects()
{
	DropTextures();
	
	while (m_pFirstUsedSurface)
	{
		TextureEntry * pVictim = m_pFirstUsedSurface;
		m_pFirstUsedSurface = pVictim->pNext;
		
		delete pVictim;
	}

	return true;
}

//*************************************************************************************
//
//*************************************************************************************
bool ITextureCache::Initialise(u32 size)
{
	size = GetNextPrime(size);
	
	m_pTextureHash = new TextureEntry *[size];
	if (m_pTextureHash == NULL)
		return false;
	
	m_dwTextureHashSize = size;
	
	for (DWORD i = 0; i < size; i++)
		m_pTextureHash[i] = NULL;
	
	return true;	// Success!
	
}

//*************************************************************************************
//
//*************************************************************************************
void	ITextureCache::GetUsedTextureStats( u32 * p_num_textures,
											u32 * p_video_memory_used,
											u32 * p_system_memory_used ) const
{
	u32	num_textures( 0 );
	u32	video_memory_used( 0 );
	u32	system_memory_used( 0 );

	for( u32 i = 0; i < m_dwTextureHashSize; ++i )
	{
		for( TextureEntry * p_te = m_pTextureHash[ i ]; p_te != NULL; p_te = p_te->pNext )
		{
			num_textures++;
			video_memory_used += p_te->GetVideoMemoryUsage();
			system_memory_used += p_te->GetSystemMemoryUsage();
		}
	}

	*p_num_textures = num_textures;
	*p_video_memory_used = video_memory_used;
	*p_system_memory_used = system_memory_used;
}

//*************************************************************************************
//
//*************************************************************************************
void	ITextureCache::GetOldTextureStats( u32 * p_num_textures,
										   u32 * p_video_memory_used,
										   u32 * p_system_memory_used ) const
{
	u32	num_textures( 0 );
	u32	video_memory_used( 0 );
	u32	system_memory_used( 0 );

	for( TextureEntry * p_te = m_pFirstUsedSurface; p_te != NULL; p_te = p_te->pNext )
	{
		num_textures++;
		video_memory_used += p_te->GetVideoMemoryUsage();
		system_memory_used += p_te->GetSystemMemoryUsage();
	}

	*p_num_textures = num_textures;
	*p_video_memory_used = video_memory_used;
	*p_system_memory_used = system_memory_used;
}

//*************************************************************************************
// Purge any textures whos last usage was over 5 seconds ago
//*************************************************************************************
void ITextureCache::PurgeOldTextures()
{
	if (m_pTextureHash == NULL)
		return;
	
	Lock();
	
	//static const DWORD dwFramesToKill = 5*30;			// 5 secs at 30 fps
	//static const DWORD dwFramesToDelete = 30*30;		// 30 secs at 30 fps
	static const DWORD dwFramesToKill = 5;
	static const DWORD dwFramesToDelete = 30;
	
	for (DWORD i = 0; i < m_dwTextureHashSize; i++)
	{
		TextureEntry * pEntry;
		TextureEntry * pNext;
		
		pEntry = m_pTextureHash[i];
		while (pEntry)
		{
			pNext = pEntry->pNext;
			
			if ( gRDPFrame - pEntry->FrameLastUsed > dwFramesToKill )
			{
				RemoveTextureEntry(pEntry);
			}
			pEntry = pNext;
		}
	}
	
	
	// Remove any old textures that haven't been recycled in 1 minute or so
	// Normally these would be reused
	TextureEntry * pPrev;
	TextureEntry * pCurr;
	TextureEntry * pNext;
	
	
	pPrev = NULL;
	pCurr = m_pFirstUsedSurface;
	
	while (pCurr)
	{
		pNext = pCurr->pNext;
		
		if ( gRDPFrame - pCurr->FrameLastUsed > dwFramesToDelete )
		{
			// Everything from this point on should be too old!
			// Remove from list
			if (pPrev != NULL) pPrev->pNext        = pCurr->pNext;
			else			   m_pFirstUsedSurface = pCurr->pNext;
			
			//DBGConsole_Msg(0, "Killing old used texture (%d x %d)", pCurr->dwWidth, pCurr->dwHeight);
			
			delete pCurr;
			
			// pPrev remains the same
			pCurr = pNext;	
		}
		else
		{
			pPrev = pCurr;
			pCurr = pNext;
		}
	}
	
	Unlock();
}

//*************************************************************************************
//
//*************************************************************************************
void ITextureCache::DropTextures()
{
	if (m_pTextureHash == NULL)
		return;
	
	Lock();
	
	DWORD dwCount = 0;
	DWORD dwTotalUses = 0;
	
	for (DWORD i = 0; i < m_dwTextureHashSize; i++)
	{
		while (m_pTextureHash[i])
		{
			TextureEntry *pTVictim = m_pTextureHash[i];
			m_pTextureHash[i] = pTVictim->pNext;
			
			dwTotalUses += pTVictim->dwUses;
			dwCount++;
			
			AddToRecycleList(pTVictim);
		}
	}
	
	DBGConsole_Msg(0, "Texture Handler: %d entries in hash table, %d uses.", dwCount, dwTotalUses);
	
	Unlock();
	
}

//*************************************************************************************
// Add to the recycle list
//*************************************************************************************
void ITextureCache::AddToRecycleList(TextureEntry *pEntry)
{
	if (pEntry->pTexture == NULL)
	{
		// No point in saving!
		delete pEntry;
	}
	else
	{
		// Add to the list
		pEntry->pNext = m_pFirstUsedSurface;
		m_pFirstUsedSurface = pEntry;
	}
}

//*************************************************************************************
// Search for a texture of the specified dimensions to recycle
//*************************************************************************************
TextureEntry * ITextureCache::ReviveUsedTexture(DWORD dwWidth, DWORD dwHeight)
{
	TextureEntry * pPrev;
	TextureEntry * pCurr;
	
	pPrev = NULL;
	pCurr = m_pFirstUsedSurface;
	
	while (pCurr)
	{
		if (pCurr->ti.Width == dwWidth &&
			pCurr->ti.Height == dwHeight)
		{
			// Remove from list
			if (pPrev != NULL) pPrev->pNext        = pCurr->pNext;
			else			   m_pFirstUsedSurface = pCurr->pNext;
			
			//DBGConsole_Msg(0, "Reviving used texture (%d x %d)", dwWidth, dwHeight);
			// Initialise any fields:
			return pCurr;
		}
		
		pPrev = pCurr;
		pCurr = pCurr->pNext;
	}

	return NULL;
}

//*************************************************************************************
//
//*************************************************************************************
DWORD ITextureCache::Hash(DWORD dwValue)
{
	// Divide by four, because most textures will be on a 4 byte boundry, so bottom four
	// bits are null
	return (dwValue>>2) % m_dwTextureHashSize;
}

//*************************************************************************************
//
//*************************************************************************************
void ITextureCache::AddTextureEntry(TextureEntry *pEntry)
{	
	DWORD dwKey = Hash(pEntry->ti.Address);
	
	if (m_pTextureHash == NULL)
		return;
		
	Lock();
	
	// Add to head (not tail, for speed - new textures are more likely to be accessed next)
	pEntry->pNext = m_pTextureHash[dwKey];
	m_pTextureHash[dwKey] = pEntry;
	
	Unlock();
	
}

//*************************************************************************************
//
//*************************************************************************************
TextureEntry * ITextureCache::GetEntry(TextureInfo * pti)
{
	TextureEntry *pEntry;
	
	if (m_pTextureHash == NULL)
		return NULL;
	
	// See if it is already in the hash table
	DWORD dwKey = Hash(pti->Address);
	for (pEntry = m_pTextureHash[dwKey]; pEntry; pEntry = pEntry->pNext)
	{
		if ( pEntry->ti.IsEqual(*pti) )
			return pEntry;
	}
	return NULL;
}

//*************************************************************************************
//
//*************************************************************************************
void ITextureCache::RemoveTextureEntry(TextureEntry * pEntry)
{
	TextureEntry * pPrev;
	TextureEntry * pCurr;
	
	if (m_pTextureHash == NULL)
		return;
	
	Lock();
	
	//DBGConsole_Msg(0, "Remove Texture entry!");
	
	// See if it is already in the hash table
	DWORD dwKey = Hash(pEntry->ti.Address);
	
	pPrev = NULL;
	pCurr = m_pTextureHash[dwKey];
	
	while (pCurr)
	{
		// Check that the attributes match
		if (pCurr->ti.IsEqual(pEntry->ti))
		{
			if (pPrev != NULL) pPrev->pNext = pCurr->pNext;
			else			   m_pTextureHash[dwKey] = pCurr->pNext;

			break;
		}
		
		pPrev = pCurr;
		pCurr = pCurr->pNext;
	}
	
	Unlock();
	
	if (pCurr == NULL)
	{
		DBGConsole_Msg(0, "Entry not found!!!");
	}
	
	AddToRecycleList(pEntry);
}

//*************************************************************************************
//
//*************************************************************************************
TextureEntry * ITextureCache::CreateEntry(TextureInfo * pti)
{
	TextureEntry * pEntry;
	
	// Find a used texture
	pEntry = ReviveUsedTexture(pti->Width, pti->Height);
	if (pEntry == NULL)
	{
		// Couldn't find on - recreate!
		pEntry = new TextureEntry;
		if (pEntry == NULL)
			return NULL;
		
		pEntry->pTexture = CTexture::Create(pti->Width, pti->Height);

		// Ignore failure for now. This prevents us trying to recreate
		// the texture each frame, which would really slow things down
	}
	
	// Initialise
	pEntry->pNext = NULL;
	pEntry->ti = *pti;
	pEntry->dwUses = 0;
	pEntry->FrameLastUsed = gRDPFrame;
	pEntry->dwCRC = CalculateCRC(pti);
	
	// Add to the hash table
	AddTextureEntry(pEntry);
	return pEntry;	
}

//*************************************************************************************
//
//*************************************************************************************
namespace
{
	u32 pixels2bytes( u32 pixels, u32 size )
	{
		return ((pixels << size)+1) / 2;
	}
}

//*************************************************************************************
//
//*************************************************************************************
DWORD ITextureCache::CalculateCRC(TextureInfo * pti) const
{
	DWORD dwCRC;
	BYTE * pStart;
	DWORD dwBytesPerLine;
	//s32 x,y;

	// If CRC checking is disabled, alwasy return 0
	if ( g_ROM.settings.bDisableTextureCRC )
		return 0;
	
	dwBytesPerLine = pixels2bytes( pti->Width, pti->Size );

	//DBGConsole_Msg(0, "BytesPerLine: %d", dwBytesPerLine);
	
	// A very simple crc - just summation
	dwCRC = 0;

	u32 offset( (pti->Top * pti->Pitch) + pixels2bytes( pti->Left, pti->Size ) ); 

	//DAEDALUS_ASSERT( (pti->Address + offset + pti->Height * pti->Pitch) < 4*1024*1024, "Address of texture is out of bounds" );

	pStart = g_pu8RamBase + pti->Address;
	pStart += offset;
	for (u32 y = 0; y < pti->Height; y++)		// Do every nth line?
	{
		// Byte fiddling won't work, but this probably doesn't matter
		dwCRC = daedalus::hash( &pStart[0], dwBytesPerLine, dwCRC );
		pStart += pti->Pitch;
	}

	if (pti->Format == G_IM_FMT_CI)
	{
		u32 bytes;
		if ( pti->Size == G_IM_SIZ_4b )		bytes = 16  * 4;
		else								bytes = 256 * 4;

		pStart = &gTextureMemory[ pti->TmemPalAddress << 3 ];
		dwCRC = daedalus::hash( &pStart[0], bytes, dwCRC );
	}

	return dwCRC;
}

//*************************************************************************************
//
//*************************************************************************************
CTexture * ITextureCache::GetTexture(TextureInfo * pti)
{
	TextureEntry * pEntry;
	
	START_PROFILE( CProfiler::PROFILE_GFX_TCACHE );
		pEntry = GetEntryAndUpdate(pti);
	STOP_PROFILE( CProfiler::PROFILE_GFX_TCACHE );
	if (pEntry == NULL)
		return NULL;

	return pEntry->pTexture;
}

//*************************************************************************************
// If already in table, return cached copy
// Otherwise, create surfaces, and load texture into memory
//*************************************************************************************
TextureEntry * ITextureCache::GetEntryAndUpdate(TextureInfo * pti)
{
	TextureEntry *pEntry;
	DWORD dwCRC;
	
	//
	// Retrieve the texture from the cache (if it already exists)
	//
	pEntry = GetEntry(pti);
	if (pEntry == NULL)
	{
		// We need to create a new entry, and add it to the hash table.
		pEntry = CreateEntry(pti);
		if (pEntry == NULL)
			return NULL;

	}
	else
	{
		// See if we've already checked that this entry is valid this frame
		if (pEntry->FrameLastUsed == gRDPFrame)
		{
			pEntry->dwUses++;
			return pEntry;
		}

		//
		// Generate and check the current crc
		//
		dwCRC = CalculateCRC(pti);

		// Check if the CRCs match, if so return. Don't do this for newly created entries
		if (pEntry->dwCRC == dwCRC)
		{
			// Tile is ok, return
			pEntry->dwUses++;
			pEntry->FrameLastUsed = gRDPFrame;			
			return pEntry;
		}

		// The crc needs updating
		pEntry->dwCRC = dwCRC;
	}

		
	//
	// The texture needs to be updated for this frame
	//
	START_PROFILE( CProfiler::PROFILE_GFX_CONVERT );
		if (pEntry->pTexture != NULL)
		{
#if RDP_EMULATE_TMEM
			DecompressTmemTexture(pEntry);		
#else
			DecompressTexture(pEntry);		
#endif
		}
	STOP_PROFILE( CProfiler::PROFILE_GFX_CONVERT );
	
	return pEntry;
	
}

//*************************************************************************************
// 
//*************************************************************************************
static const char *pszImgFormat[8] = {"RGBA", "YUV", "CI", "IA", "I", "?1", "?2", "?3"};
static const BYTE pnImgSize[4]   = {4, 8, 16, 32};

void ITextureCache::DecompressTexture(TextureEntry * pEntry)
{
	
	BOOL bHandled;	
	
	//void * pSrc = g_pu8RamBase + pEntry->ti.Address;
	
	static DWORD dwCount = 0;
	
	bHandled = FALSE;
	
	switch (pEntry->ti.Format)
	{
	case G_IM_FMT_RGBA:
		switch (pEntry->ti.Size)
		{
		case G_IM_SIZ_16b:
			ConvertRGBA16(pEntry->pTexture, pEntry->ti);
			bHandled = TRUE;
			break;
		case G_IM_SIZ_32b:
			ConvertRGBA32(pEntry->pTexture, pEntry->ti);
			bHandled = TRUE;
			break;
		}
		break;
		
	case G_IM_FMT_YUV:
		break;

	case G_IM_FMT_CI:
		switch (pEntry->ti.Size)
		{
		case G_IM_SIZ_4b: // 4bpp
			switch (pEntry->ti.TLutFmt)
			{
				//case G_TT_NONE:
			case G_TT_RGBA16:
				ConvertCI4_RGBA16(pEntry->pTexture, pEntry->ti);					
				bHandled = TRUE;
				break;
			case G_TT_IA16:
				ConvertCI4_IA16(pEntry->pTexture, pEntry->ti);					
				bHandled = TRUE;
				break;
			}
			break;
			
		case G_IM_SIZ_8b: // 8bpp
			switch(pEntry->ti.TLutFmt)
			{
			case G_TT_RGBA16:
				ConvertCI8_RGBA16(pEntry->pTexture, pEntry->ti);					
				bHandled = TRUE;
				break;
			case G_TT_IA16:
				ConvertCI8_IA16(pEntry->pTexture, pEntry->ti);					
				bHandled = TRUE;
				break;
			}
			break;
		}
		
		break;

	case G_IM_FMT_IA:
		switch (pEntry->ti.Size)
		{
		case G_IM_SIZ_4b:
			ConvertIA4(pEntry->pTexture, pEntry->ti);
			bHandled = TRUE;
			break;
		case G_IM_SIZ_8b:
			ConvertIA8(pEntry->pTexture, pEntry->ti);
			bHandled = TRUE;
			break;
		case G_IM_SIZ_16b:
			ConvertIA16(pEntry->pTexture, pEntry->ti);
			bHandled = TRUE;
			break;
		case G_IM_SIZ_32b:
			break;
		}
		break;

	case G_IM_FMT_I:
		switch (pEntry->ti.Size)
		{
		case G_IM_SIZ_4b:
			ConvertI4(pEntry->pTexture, pEntry->ti);
			bHandled = TRUE;
			break;
		case G_IM_SIZ_8b:
			ConvertI8(pEntry->pTexture, pEntry->ti);
			bHandled = TRUE;
			break;
		}
		break;
	default:
		break;
	}
	
	if (bHandled)
	{
#ifdef LIBPNG_SUPPORT
		if ( m_bDumpTextures )
		{
			
			TCHAR szFileName[MAX_PATH+1];
			TCHAR szFilePath[MAX_PATH+1];
			TCHAR szDumpDir[MAX_PATH+1];
			
			wsprintf(szDumpDir, "%s\\%s", TEXT("tile_dumps"), g_ROM.szGameName);
			
			Dump_GetDumpDirectory(szFilePath, szDumpDir);
			
			wsprintf(szFileName, TEXT("%08.8x-%s_%dbpp-%dx%d-%dx%d.png"),
				pEntry->ti.Address, pszImgFormat[pEntry->ti.Format], pnImgSize[pEntry->ti.Size],
				pEntry->ti.Left, pEntry->ti.Top,
				pEntry->ti.Width,pEntry->ti.Height);
			
			PathAppend(szFilePath, szFileName);
			
			pEntry->pTexture->DumpImageAsPNG(szFilePath, pEntry->ti.Width,pEntry->ti.Height);
		}
#endif	
		
	}
	else
	{
		static BOOL bWarningEmitted = FALSE;
		
		if (!bWarningEmitted)
		{			
			DBGConsole_Msg(0, "DecompressTexture: Unable to decompress %s/%dbpp", pszImgFormat[pEntry->ti.Format], pnImgSize[pEntry->ti.Size]);	
			//bWarningEmitted = TRUE;
		}
	}
	
	
	dwCount++;	
}

//*************************************************************************************
// Load a texture from tmem
//*************************************************************************************
void ITextureCache::DecompressTmemTexture(TextureEntry * pEntry)
{
	
	BOOL bHandled;	
	
	static DWORD dwCount = 0;
	
	bHandled = FALSE;
	
	switch (pEntry->ti.Format)
	{
	case G_IM_FMT_RGBA:
		switch (pEntry->ti.Size)
		{
		case G_IM_SIZ_16b:
			ConvertTmemRGBA16(pEntry->pTexture, pEntry->ti);
			bHandled = TRUE;
			break;
		case G_IM_SIZ_32b:
			ConvertTmemRGBA32(pEntry->pTexture, pEntry->ti);
			bHandled = TRUE;
			break;
		}
		break;
		
	case G_IM_FMT_YUV:
		break;

	case G_IM_FMT_CI:
		switch (pEntry->ti.Size)
		{
		case G_IM_SIZ_4b: // 4bpp
			switch (pEntry->ti.TLutFmt)
			{
				//case G_TT_NONE:
			case G_TT_RGBA16:
				ConvertTmemCI4_RGBA16(pEntry->pTexture, pEntry->ti);					
				bHandled = TRUE;
				break;
			case G_TT_IA16:
				ConvertTmemCI4_IA16(pEntry->pTexture, pEntry->ti);					
				bHandled = TRUE;
				break;
			}
			break;
			
		case G_IM_SIZ_8b: // 8bpp
			switch(pEntry->ti.TLutFmt)
			{
			case G_TT_RGBA16:
				ConvertTmemCI8_RGBA16(pEntry->pTexture, pEntry->ti);					
				bHandled = TRUE;
				break;
			case G_TT_IA16:
				ConvertTmemCI8_IA16(pEntry->pTexture, pEntry->ti);					
				bHandled = TRUE;
				break;
			}
			break;
		}
		
		break;

	case G_IM_FMT_IA:
		switch (pEntry->ti.Size)
		{
		case G_IM_SIZ_4b:
			ConvertTmemIA4(pEntry->pTexture, pEntry->ti);
			bHandled = TRUE;
			break;
		case G_IM_SIZ_8b:
			ConvertTmemIA8(pEntry->pTexture, pEntry->ti);
			bHandled = TRUE;
			break;
		case G_IM_SIZ_16b:
			ConvertTmemIA16(pEntry->pTexture, pEntry->ti);
			bHandled = TRUE;
			break;
		case G_IM_SIZ_32b:
			break;
		}
		break;

	case G_IM_FMT_I:
		switch (pEntry->ti.Size)
		{
		case G_IM_SIZ_4b:
			ConvertTmemI4(pEntry->pTexture, pEntry->ti);
			bHandled = TRUE;
			break;
		case G_IM_SIZ_8b:
			ConvertTmemI8(pEntry->pTexture, pEntry->ti);
			bHandled = TRUE;
			break;
		}
		break;
	default:
		break;
	}
	
	if (bHandled)
	{
#ifdef LIBPNG_SUPPORT
		if ( m_bDumpTextures )
		{
			
			TCHAR szFileName[MAX_PATH+1];
			TCHAR szFilePath[MAX_PATH+1];
			TCHAR szDumpDir[MAX_PATH+1];
			
			wsprintf(szDumpDir, "%s\\%s", TEXT("tile_dumps"), g_ROM.szGameName);
			
			Dump_GetDumpDirectory(szFilePath, szDumpDir);
			
			wsprintf(szFileName, TEXT("%08.8x-%s_%dbpp-%dx%d-%dx%d.png"),
				pEntry->ti.Address, pszImgFormat[pEntry->ti.Format], pnImgSize[pEntry->ti.Size],
				pEntry->ti.Left, pEntry->ti.Top,
				pEntry->ti.Width,pEntry->ti.Height);
			
			PathAppend(szFilePath, szFileName);
			
			pEntry->pTexture->DumpImageAsPNG(szFilePath, pEntry->ti.Width,pEntry->ti.Height);
		}
#endif	
		
	}
	else
	{
		static BOOL bWarningEmitted = FALSE;
		
		if (!bWarningEmitted)
		{			
			DBGConsole_Msg(0, "DecompressTexture: Unable to decompress %s/%dbpp", pszImgFormat[pEntry->ti.Format], pnImgSize[pEntry->ti.Size]);	
			//bWarningEmitted = TRUE;
		}
	}
	
	
	dwCount++;	
}

//*************************************************************************************
//	Display info on the number of textures in the cache
//*************************************************************************************
void	ITextureCache::DisplayStats()
{
	u32		used_textures;
	u32		used_texture_video_mem;
	u32		used_texture_system_mem;
	u32		old_textures;
	u32		old_texture_video_mem;
	u32		old_texture_system_mem;

	Daedalus_ForceDebugScreenToFront();

	GetUsedTextureStats( &used_textures, &used_texture_video_mem, &used_texture_system_mem );
	GetOldTextureStats( &old_textures, &old_texture_video_mem, &old_texture_system_mem );

	pspDebugScreenPrintf( "Texture Cache Stats\n" );
	pspDebugScreenPrintf( " Used: %3d v:%4dKB s:%4dKB\n", used_textures, used_texture_video_mem / 1024, used_texture_system_mem / 1024 );
	pspDebugScreenPrintf( "  Old: %3d v:%4dKB s:%4dKB\n", old_textures, old_texture_video_mem / 1024, old_texture_system_mem / 1024 );
}
